"""This module contains helper functions when working with sklearn (scikit-learn) objects;
in particular, for evaluating models"""
# pylint: disable=too-many-lines
import math
import warnings
from typing import Tuple, Union, Optional

import numpy as np
import pandas as pd
import scipy.stats as st
import seaborn as sns
from matplotlib import pyplot as plt
from pandas.io.formats.style import Styler
from sklearn.metrics import confusion_matrix, roc_auc_score, r2_score
from sklearn.model_selection._search import BaseSearchCV  # noqa

import helpsk.color as hcolor
import helpsk.pandas_style as hstyle
import helpsk.string as hstring
# pylint: disable=too-many-locals
from helpsk.exceptions import HelpskParamValueError
from helpsk.plot import STANDARD_WIDTH_HEIGHT
from helpsk.validation import assert_is_close, assert_true

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=Warning)
    from statsmodels import api as sm  # https://github.com/statsmodels/statsmodels/issues/3814


# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-public-methods
class SearchCVParser:
    """
    This class contains the logic to parse and extract information from a BaseSearchCV object (e.g.
    GridSearchCV, RandomizedSearchCV, BayesSearchCV)
    """
    def __init__(self,
                 searcher: BaseSearchCV,
                 higher_score_is_better: bool = True,
                 new_param_column_names: Union[dict, None] = None):

        self._higher_score_is_better = higher_score_is_better
        self.num_repeats = searcher.cv.n_repeats
        self.num_folds = searcher.cv.cvargs['n_splits']

        self._results, self._primary_score_standard_errors = \
            SearchCVParser.cv_results_to_dataframe(searcher=searcher,
                                                   higher_score_is_better=higher_score_is_better)

        if new_param_column_names:
            self._results.rename(columns=new_param_column_names, inplace=True)

        # mean/std times per trial; i.e. to get total time, multiple by number_of_runs_per_trial
        self.mean_fit_time = searcher.cv_results_['mean_fit_time']
        self.std_fit_time = searcher.cv_results_['std_fit_time']
        self.mean_score_time = searcher.cv_results_['mean_score_time']
        self.std_score_time = searcher.cv_results_['std_score_time']

        assert_true(len(self.mean_fit_time) == self.number_of_trials)
        assert_is_close(self.total_time, self.mean_time_per_trial * self.number_of_trials)

    @property
    def number_of_trials(self) -> int:
        """"A single trial contains the cross validation runs for a single set of hyper-parameters. The
        'number of trials' is basically the number of combinations of different hyper-parameters that were
        cross validated."""
        return self.results.shape[0]

    @property
    def number_of_runs_per_trial(self) -> int:
        """This is the number of CV folds multiplied by the number of CV repeats."""
        return self.num_folds * self.num_repeats

    @property
    def fit_time_per_trial(self) -> np.array:
        """Average fit time for each trial multiplied by the number of runs per trial

        Returns:
            array containing the fit time for each trail
        """
        return self.mean_fit_time * self.number_of_runs_per_trial

    @property
    def fit_time_total(self) -> float:
        """Total fit time across all trials."""
        return float(np.sum(self.fit_time_per_trial))

    @property
    def score_time_per_trial(self):
        """Average score time for each trial multiplied by the number of runs per trial

        Returns:
            array containing the score time for each trail
        """
        return self.mean_score_time * self.number_of_runs_per_trial

    @property
    def score_time_total(self) -> float:
        """Total score time across all trials."""
        return float(np.sum(self.score_time_per_trial))

    @property
    def mean_time_per_trial(self) -> float:
        """Average time per trial"""
        return float(np.mean(self.fit_time_per_trial + self.score_time_per_trial))

    @property
    def total_time(self) -> float:
        """Total time it took across all trials"""
        return self.fit_time_total + self.score_time_total

    @property
    def score_names(self) -> list:
        """Returns a list of the names of the scores"""
        def replace_type(value):
            return value.replace(' Mean', '').replace(' 95CI.LO', '').replace(' 95CI.HI', '')
        names = [replace_type(x) for x in self.score_columns]
        # use .fromkeys to dedupe
        return list(dict.fromkeys(names))
    
    @property
    def trial_labels(self) -> pd.Series:
        """A trial is a set of hyper-parameters that were cross validated. (Each row of the `results`
         DataFrame corresponds to a single trial. The corresponding label for each trial is a single string
         containing all of the hyper-parameter names and values in the format of
         `{param1: value1, param2: value2}`.

        Returns:
            a pd.Series the same length as `number_of_trials` containing a str
        """
        def create_hyper_param_labels(data_row) -> list:
            """Creates a list of strings that represent the name/value pair for each hyper-parameter."""
            return [f"{x}: {data_row[x]}" for x in self.parameter_columns]

        def create_trial_label(data_row) -> str:
            return f"{{{hstring.collapse(create_hyper_param_labels(data_row), separate=', ')}}}"

        return self.results.apply(create_trial_label, axis=1)

    @property
    def number_of_scores(self) -> int:
        """The number of scores passed to the SearchCV object"""
        return len(self.score_names)

    @property
    def primary_score_name(self) -> str:
        """The first scorer passed to the SearchCV will be treated as the primary score."""
        return self.score_names[0]

    @property
    def primary_score_standard_errors(self) -> pd.Series:
        """The standard errors associated with the scores. Sorted to correspond to `results` DataFrame."""
        return self._primary_score_standard_errors

    @property
    def top_score(self):
        """The top 'primary' score (i.e. the average cross validation score associated with the best set of
        hyper-params."""
        return self.results.iloc[0, 0]

    @property
    def result_indexes_within_1_standard_error(self):
        """Returns the indexes of the `results` DataFrame where the primary scores (i.e. first scorer
        passed to SearchCV object; i.e. first column of the results DataFrame) are within 1 standard error of
        the highest primary score."""
        if self._higher_score_is_better:
            return self.results.index[self.results.iloc[:, 0] >=
                                      self.top_score - self.top_score_standard_error]

        return self.results.index[self.results.iloc[:, 0] <= self.top_score + self.top_score_standard_error]

    @property
    def top_score_standard_error(self) -> float:
        """The standard error associated with the top 'primary' score."""
        return self.primary_score_standard_errors.iloc[0]  # index of highest score

    @property
    def score_columns(self) -> list:
        """Returns a list of the names of the score columns"""
        return [x for x in self.results.columns[self.results.columns.str.endswith((' Mean',
                                                                                   ' 95CI.LO',
                                                                                   ' 95CI.HI'))]
                if x not in self.training_score_columns]

    @property
    def training_score_columns(self) -> list:
        """Returns a list of the names of the training score columns"""
        return self.results.columns[self.results.columns.str.contains(' Training ')].tolist()

    @property
    def parameter_columns(self) -> list:
        """Returns a list of the names of the columns associated with the hyper-parameters"""
        return [x for x in self.results.columns
                if x not in self.score_columns + self.training_score_columns]

    @property
    def results(self):
        """Main DataFrame summarizing the results of the SearchCV object."""
        return self._results.copy()

    def formatted_results(self,
                          round_by: int = 3,
                          return_train_score: bool = True,
                          exclude_no_variance_params: bool = True,
                          return_style: bool = True) -> Union[pd.DataFrame, Styler]:
        """Returns a summary of the results, either as a Data.Frame, or Data.Frame.Styler.

        If a Styler is returned, then Hyper-Parameter columns will be highlighted in blue where the primary
        score (i.e. first column) for the iteration (i.e. the row i.e. the combination of hyper-parameters
        that were cross validated) is within 1 standard error of the top primary score (i.e. first column
        first row).

        Args:
            round_by:
                the number of digits to round by for the score columns (does not round the parameter columns)
            return_train_score:
                if True, return the columns associated with the training scores, if any.
            exclude_no_variance_params
                if True, exclude columns that only have 1 unique value
            return_style:
                If True, return Styler object

        Returns:
            either a DataFrame or Styler object.
        """
        results = self.results

        if exclude_no_variance_params:
            columns_to_drop = [x for x in self.parameter_columns if len(self.results[x].unique()) == 1]
            results = results.drop(columns=columns_to_drop)

        results = results.round(dict(zip(self.score_columns + self.training_score_columns,
                                         [round_by] * len(self.score_columns + self.training_score_columns))))

        if not return_train_score and len(self.training_score_columns) > 0:
            results = results.drop(columns=self.training_score_columns)

        final_columns = results.columns  # save for style logic

        if return_style:
            results = results.style

            # note that we are not adding styles to the training scores; training scores are secondary info
            for score in self.score_names:
                mean_key = score + ' Mean'
                ci_low_key = score + ' 95CI.LO'
                ci_high_key = score + ' 95CI.HI'

                results. \
                    bar(subset=[mean_key], color=hcolor.Colors.PIGMENT_GREEN.value). \
                    bar(subset=[ci_high_key], color=hcolor.GRAY). \
                    pipe(hstyle.bar_inverse, subset=[ci_low_key], color=hcolor.GRAY). \
                    pipe(hstyle.format, round_by=round_by, hide_index=True)

            # highlight iterations whose primary score (i.e. first column of `results` dataframe) is within
            # 1 standard error of the top primary score (i.e. first column first row).
            def highlight_cols(s):
                return 'background-color: %s' % hcolor.Colors.PASTEL_BLUE.value

            # we might have removed columns (e.g. that don't have any variance) so check that the columns
            # were in the final set
            columns_to_highlight = [x for x in self.parameter_columns if x in final_columns]
            results.applymap(highlight_cols,
                             subset=pd.IndexSlice[self.result_indexes_within_1_standard_error,
                                                  columns_to_highlight])

        return results

    # pylint: disable=too-many-statements
    @staticmethod
    def cv_results_to_dataframe(searcher: BaseSearchCV,
                                higher_score_is_better: bool = True) -> tuple[pd.DataFrame, pd.Series]:
        """
        Args:
            searcher:
                A `BaseSearchCV` object that has either used a string passed to the `scoring` parameter of the
                constructor (e.g. `GridSearchCV(..., scoring='auc', ...)` or a dictionary with metric names as
                keys and callables a values.

                An example of the dictionary option:

                    scores = {
                        'ROC/AUC': SCORERS['roc_auc'],
                        'F1': make_scorer(f1_score, greater_is_better=True),
                        'Pos. Pred. Val': make_scorer(precision_score, greater_is_better=True),
                        'True Pos. Rate': make_scorer(recall_score, greater_is_better=True),
                    }
                    grid_search = GridSearchCV(..., scoring=scores, ...)
            higher_score_is_better:
                if True, higher scores are better; if False, lower scores are better
                False assumes that the scores returned from sklearn are negative and will multiple the values
                by 1.

        Returns:
            DataFrame with a row corresponding to each hyper-parameter combination a.k.a 'trial'
        """
        num_folds = searcher.cv.cvargs['n_splits']
        num_repeats = searcher.cv.n_repeats
        sample_size = num_folds * num_repeats
        results = None
        cv_results = searcher.cv_results_

        str_score_name = None
        if isinstance(searcher.scoring, dict):
            score_names = list(searcher.scoring.keys())
        elif isinstance(searcher.scoring, str):
            score_names = ['score']
            str_score_name = searcher.scoring
        else:
            mess = 'The `searcher` does not have a string or dictionary .scoring property. Cannot extract ' \
                   'scores.'
            raise HelpskParamValueError(mess)

        # extract mean and standard deviation for each score
        # if a string was passed into the searcher `scoring` parameter (e.g. 'roc/auc') then the score name
        # will just be 'mean_test_score' but we will convert it to `roc/auc Mean`. Same for standard
        # deviation. In that case, then after we are done, we need to change `score_names = ['score']` to
        # `score_names = [str_score_name]`
        for score in score_names:
            score_name = score if str_score_name is None else str_score_name
            mean_scores = cv_results['mean_test_' + score]
            mean_scores = mean_scores if higher_score_is_better else mean_scores * -1

            results = pd.concat([
                results,
                pd.DataFrame({
                    score_name + " Mean": mean_scores,
                    score_name + " St. Dev": cv_results['std_test_' + score],
                })
            ], axis=1)

            if 'mean_train_' + score in cv_results:
                mean_training_scores = cv_results['mean_train_' + score]
                mean_training_scores = mean_training_scores \
                    if higher_score_is_better else mean_training_scores * -1
                results = pd.concat([
                    results,
                    pd.DataFrame({
                        score_name + " Training Mean": mean_training_scores,
                        score_name + " Training St. Dev": cv_results['std_train_' + score],
                    })
                ], axis=1)

        # see comment above
        if str_score_name is not None:
            score_names = [str_score_name]

        def add_confidence_interval(score_name, dataframe):  # noqa
            mean_column_name = score_name + ' Mean'
            st_dev_column_name = score_name + ' St. Dev'
            score_means = dataframe[mean_column_name]
            score_standard_errors = dataframe[st_dev_column_name] / math.sqrt(sample_size)

            confidence_intervals = st.t.interval(alpha=0.95,  # confidence interval
                                                 df=sample_size - 1,  # degrees of freedom
                                                 loc=score_means,
                                                 scale=score_standard_errors)

            dataframe = dataframe.drop(columns=st_dev_column_name)

            insertion_index = dataframe.columns.get_loc(mean_column_name) + 1
            dataframe.insert(loc=insertion_index, column=score_name + ' 95CI.HI',
                             value=confidence_intervals[1])
            dataframe.insert(loc=insertion_index, column=score_name + ' 95CI.LO',
                             value=confidence_intervals[0])

            return dataframe, score_standard_errors

        primary_score_standard_errors = None
        # for each score, calculate the 95% confidence interval for the mean
        for score in score_names:
            (results, standard_errors) = add_confidence_interval(score_name=score,
                                                                 dataframe=results)

            if score == score_names[0]:  # if score is the first/primary score (or only)
                primary_score_standard_errors = standard_errors

            results[f'{score} Mean'] = results[f'{score} Mean']
            results[f'{score} 95CI.LO'] = results[f'{score} 95CI.LO']
            results[f'{score} 95CI.HI'] = results[f'{score} 95CI.HI']

            # if there are training scores, then drop the St. Dev, and don't add confidence intervals,
            # which would add too many columns (noise) and not that interesting.
            if score + ' Training Mean' in results.columns:
                results[f'{score} Training Mean'] = results[f'{score} Training Mean']
                results = results.drop(columns=score + ' Training St. Dev')

        assert_true(primary_score_standard_errors is not None)

        parameter_dataframe = pd.DataFrame(cv_results["params"])
        parameter_dataframe.columns = [x.replace('__', ' | ') for x in parameter_dataframe.columns]

        for column in parameter_dataframe.columns:
            # this will convert objects (like Transformers) into strings, which makes further operations on
            # the column (e.g. aggregations/group_by/etc) much easier.
            if parameter_dataframe[column].dtype.name == 'object':
                parameter_dataframe[column] = parameter_dataframe[column].astype(str)

        results = pd.concat([
            results,
            parameter_dataframe,
        ], axis=1)

        results = results.sort_values(by=str(list(score_names)[0]) + ' Mean',
                                      ascending=not higher_score_is_better)

        primary_score_standard_errors = primary_score_standard_errors.reindex(results.index)

        return results, primary_score_standard_errors


# pylint: disable=too-many-instance-attributes,too-many-public-methods
class TwoClassEvaluator:
    """This class calculates various metrics for Two Class (i.e. 0's/1's) prediction scenarios."""
    def __init__(self,
                 actual_values: np.ndarray,
                 predicted_scores: np.ndarray,
                 labels: Tuple[str, str],
                 score_threshold: float = 0.5
                 ):
        """
        Args:
            actual_values:
                array of 0's and 1's
            predicted_scores:
                array of values from `predict_proba()`; NOT the actual labels
            labels:
                tuple containing the label of the negative class in the first index and the positive class
                in the second index
            score_threshold:
                the score/probability threshold for turning scores into 0's and 1's and corresponding labels
        """

        assert len(actual_values) == len(predicted_scores)

        if not all(np.unique(actual_values) == [0, 1]):
            message = f"Values of `actual_values` should 0 or 1. Found `{np.unique(actual_values)}`"
            raise HelpskParamValueError(message)

        if not all(np.logical_and(predicted_scores >= 0, predicted_scores <= 1)):
            message = "Values of `predicted_scores` should be between 0 and 1."
            raise HelpskParamValueError(message)

        if not any(np.logical_and(predicted_scores > 0, predicted_scores < 1)):
            message = "Values of `predicted_scores` should not all be 0's and 1's."
            raise HelpskParamValueError(message)

        self._labels = labels
        self._actual_values = actual_values
        self._predicted_scores = predicted_scores
        self.score_threshold = score_threshold
        predicted_values = np.where(predicted_scores > self.score_threshold, 1, 0)
        self._confusion_matrix = confusion_matrix(y_true=actual_values, y_pred=predicted_values)

        self.sample_size = len(actual_values)
        assert self.sample_size == self._confusion_matrix.sum()

        true_negatives, false_positives, false_negatives, true_positives = self._confusion_matrix.ravel()

        self._actual_positives = true_positives + false_negatives
        assert self._actual_positives == sum(self._actual_values == 1)

        self._actual_negatives = true_negatives + false_positives

        self._true_negatives = true_negatives
        self._false_positives = false_positives
        self._false_negatives = false_negatives
        self._true_positives = true_positives

        self.auc = roc_auc_score(y_true=actual_values, y_score=predicted_scores)

    @property
    def true_positive_rate(self) -> float:
        """True Positive Rate"""
        return 0 if self._actual_positives == 0 else self._true_positives / self._actual_positives

    @property
    def true_negative_rate(self) -> float:
        """True Negative Rate i.e. Specificity"""
        return 0 if self._actual_negatives == 0 else self._true_negatives / self._actual_negatives

    @property
    def false_negative_rate(self) -> float:
        """False Negative Rate"""
        return 0 if self._actual_positives == 0 else self._false_negatives / self._actual_positives

    @property
    def false_positive_rate(self) -> float:
        """False Positive Rate"""
        return 0 if self._actual_negatives == 0 else self._false_positives / self._actual_negatives

    @property
    def accuracy(self) -> Union[float, None]:
        """accuracy"""
        return None if self.sample_size == 0 else \
            (self._true_negatives + self._true_positives) / self.sample_size

    @property
    def error_rate(self) -> Union[float, None]:
        """error_rate"""
        return None if self.sample_size == 0 else \
            (self._false_positives + self._false_negatives) / self.sample_size

    @property
    def positive_predictive_value(self) -> float:
        """Positive Predictive Value i.e. Precision"""
        return 0 if (self._true_positives + self._false_positives) == 0 else \
            self._true_positives / (self._true_positives + self._false_positives)

    @property
    def negative_predictive_value(self) -> float:
        """Negative Predictive Value"""
        return 0 if (self._true_negatives + self._false_negatives) == 0 else \
            self._true_negatives / (self._true_negatives + self._false_negatives)

    @property
    def prevalence(self) -> Union[float, None]:
        """Prevalence"""
        return None if self.sample_size == 0 else \
            self._actual_positives / self.sample_size

    @property
    def kappa(self) -> Union[float, None]:
        """Kappa"""
        if self.sample_size == 0 or \
                ((self._true_negatives + self._false_negatives) / self.sample_size) == 0:
            return None
        # proportion of the actual agreements
        # add the proportion of all instances where the predicted type and actual type agree
        pr_a = (self._true_negatives + self._true_positives) / self.sample_size
        # probability of both predicted and actual being negative
        p_negative_prediction_and_actual = \
            ((self._true_negatives + self._false_positives) / self.sample_size) * \
            ((self._true_negatives + self._false_negatives) / self.sample_size)
        # probability of both predicted and actual being positive
        p_positive_prediction_and_actual = \
            self.prevalence * ((self._false_positives + self._true_positives) / self.sample_size)
        # probability that chance alone would lead the predicted and actual values to match, under the
        # assumption that both are selected randomly (i.e. implies independence) according to the observed
        # proportions (probability of independent events = P(A & B) == P(A) * P(B)
        pr_e = p_negative_prediction_and_actual + p_positive_prediction_and_actual
        return (pr_a - pr_e) / (1 - pr_e)

    @property
    def f1_score(self) -> float:
        """F1 Score
        https://en.wikipedia.org/wiki/F-score
        """
        return self.fbeta_score(beta=1)

    def fbeta_score(self, beta: float) -> float:
        """
        :param beta: The `beta` parameter determines the weight of precision in the combined score.
            `beta < 1` lends more weight to precision (i.e. positive predictive value), while
            `beta > 1` favors recall (i.e. true positive rate)
            (`beta -> 0` considers only precision, `beta -> inf` only recall).
            http://scikit-learn.org/stable/modules/generated/sklearn.metrics.fbeta_score.html
        :return:
        """
        if self.positive_predictive_value is None or self.sensitivity is None or \
                (self.positive_predictive_value + self.sensitivity) == 0:
            return 0

        return (1 + (beta ** 2)) * (self.positive_predictive_value * self.sensitivity) / \
               (((beta ** 2) * self.positive_predictive_value) + self.sensitivity)

    @property
    def sensitivity(self) -> float:
        """Sensitivity i.e. True Positive Rate"""
        return self.true_positive_rate

    @property
    def specificity(self) -> float:
        """Specificity i.e. True Negative Rate"""
        return self.true_negative_rate

    @property
    def precision(self) -> float:
        """Precision i.e. Positive Predictive Value"""
        return self.positive_predictive_value

    @property
    def recall(self):
        """Recall i.e. True Positive Rate"""
        return self.true_positive_rate

    @property
    def all_metrics(self) -> dict:
        """All of the metrics are returned as a dictionary."""

        auc_message = 'Area under the ROC curve (true pos. rate vs false pos. rate); ' \
                      'ranges from 0.5 (purely random classifier) to 1.0 (perfect classifier)'

        tpr_message = f'{self.true_positive_rate:.1%} of positive instances were correctly identified.; ' \
                      f'i.e. {self._true_positives} "{self._labels[1]}" labels were correctly identified ' \
                      f'out of {self._actual_positives} instances; a.k.a Sensitivity/Recall'

        tnr_message = f'{self.true_negative_rate:.1%} of negative instances were correctly identified.; ' \
                      f'i.e. {self._true_negatives} "{self._labels[0]}" labels were correctly identified ' \
                      f'out of {self._actual_negatives} instances'

        fpr_message = f'{self.false_positive_rate:.1%} of negative instances were incorrectly identified ' \
                      f'as positive; ' \
                      f'i.e. {self._false_positives} "{self._labels[0]}" labels were incorrectly ' \
                      f'identified as "{self._labels[1]}", out of {self._actual_negatives} instances'

        fnr_message = f'{self.false_negative_rate:.1%} of positive instances were incorrectly identified ' \
                      f'as negative; ' \
                      f'i.e. {self._false_negatives} "{self._labels[1]}" labels were incorrectly ' \
                      f'identified as "{self._labels[0]}", out of {self._actual_positives} instances'

        ppv_message = f'When the model claims an instance is positive, it is correct ' \
                      f'{self.positive_predictive_value:.1%} of the time; ' \
                      f'i.e. out of the {self._true_positives + self._false_positives} times the model ' \
                      f'predicted "{self._labels[1]}", it was correct {self._true_positives} ' \
                      f'times; a.k.a precision'

        npv_message = f'When the model claims an instance is negative, it is correct ' \
                      f'{self.negative_predictive_value:.1%} of the time; ' \
                      f'i.e. out of the {self._true_negatives + self._false_negatives} times the model ' \
                      f'predicted "{self._labels[0]}", it was correct {self._true_negatives} times'

        f1_message = 'The F1 score can be interpreted as a weighted average of the precision and recall, ' \
                     'where an F1 score reaches its best value at 1 and worst score at 0.'

        accuracy_message = f'{self.accuracy:.1%} of instances were correctly identified'
        error_message = f'{self.error_rate:.1%} of instances were incorrectly identified'
        prevalence_message = f'{self.prevalence:.1%} of the data are positive; i.e. out of ' \
                             f'{self.sample_size} total observations; {self._actual_positives} are labeled ' \
                             f'as "{self._labels[1]}"'
        total_obs_message = f'There are {self.sample_size} total observations; i.e. sample size'

        return {'AUC': (self.auc, auc_message),
                'True Positive Rate': (self.true_positive_rate, tpr_message),
                'True Negative Rate': (self.true_negative_rate, tnr_message),
                'False Positive Rate': (self.false_positive_rate, fpr_message),
                'False Negative Rate': (self.false_negative_rate, fnr_message),
                'Positive Predictive Value': (self.positive_predictive_value, ppv_message),
                'Negative Predictive Value': (self.negative_predictive_value, npv_message),
                'F1 Score': (self.f1_score, f1_message),
                'Accuracy': (self.accuracy, accuracy_message),
                'Error Rate': (self.error_rate, error_message),
                '% Positive': (self.prevalence, prevalence_message),
                'Total Observations': (self.sample_size, total_obs_message)}

    def all_metrics_df(self,
                       return_details: bool = True,
                       return_style: bool = False,
                       round_by: Optional[int] = None) -> Union[pd.DataFrame, Styler]:
        """All of the metrics are returned as a DataFrame.

        Args:
            return_details:
                if True, then return descriptions of score and more information in an additional column
            return_style:
                if True, return styler object; else return dataframe
            round_by:
                the number of digits to round by; if None, then don't round
        """
        if return_details:
            result = pd.DataFrame.from_dict(self.all_metrics,
                                            orient='index',
                                            columns=['Scores', 'Details'])
        else:
            result = pd.DataFrame.from_dict({key: value[0] for key, value in self.all_metrics.items()},
                                            orient='index',
                                            columns=['Scores'])

        if round_by:
            result['Scores'] = result['Scores'].round(round_by)

        if return_style:
            subset_scores = [x for x in result.index.values if x != 'Total Observations']

            subset_scores = pd.IndexSlice[result.loc[subset_scores, :].index, 'Scores']
            subset_negative_bad = pd.IndexSlice[result.loc[['False Positive Rate',
                                                            'False Negative Rate'], 'Scores'].index, 'Scores']
            subset_secondary = pd.IndexSlice[result.loc[['Accuracy', 'Error Rate', '% Positive'],
                                                        'Scores'].index, 'Scores']
            subset_total_observations = pd.IndexSlice[result.loc[['Total Observations'],
                                                                 'Scores'].index, 'Scores']

            result = result.style

            if round_by:
                result = result.format(precision=round_by)

            result = result.format(subset=subset_total_observations,
                                   thousands=',',
                                   precision=0)

            result = result. \
                bar(subset=subset_scores, color=hcolor.Colors.PIGMENT_GREEN.value, vmin=0, vmax=1). \
                bar(subset=subset_negative_bad, color=hcolor.Colors.POPPY.value, vmin=0, vmax=1). \
                bar(subset=subset_secondary, color=hcolor.GRAY, vmin=0, vmax=1)

        return result

    def plot_confusion_matrix(self):
        """Plots a heatmap of the confusion matrix."""
        labels = np.array([[f'True Negatives\n{self._true_negatives}\n{self._true_negatives / self.sample_size:.1%}',  # pylint: disable=line-too-long  # noqa
                            f'False Positives\n{self._false_positives}\n{self._false_positives / self.sample_size:.1%}'],  # pylint: disable=line-too-long  # noqa
                           [f'False Negatives\n{self._false_negatives}\n{self._false_negatives / self.sample_size:.1%}',  # pylint: disable=line-too-long  # noqa
                            f'True Positives\n{self._true_positives}\n{self._true_positives / self.sample_size:.1%}']])  # pylint: disable=line-too-long  # noqa

        axis = plt.subplot()
        sns.heatmap(self._confusion_matrix, annot=labels, cmap='Blues', ax=axis, fmt='')
        # labels, title and ticks
        axis.set_xlabel('Predicted')
        axis.set_ylabel('Actual')
        # axis.set_title('Confusion Matrix');
        axis.xaxis.set_ticklabels([self._labels[0], self._labels[1]])
        axis.yaxis.set_ticklabels([self._labels[0], self._labels[1]])
        plt.tight_layout()

    def plot_auc_curve(self, figure_size: tuple = STANDARD_WIDTH_HEIGHT):
        """Plots the ROC AUC

        Args:
            figure_size:
                tuple containing `(width, height)` of plot. The default height is defined by
                `helpsk.plot.STANDARD_HEIGHT`, and the default width is
                `helpsk.plot.STANDARD_HEIGHT / helpsk.plot.GOLDEN_RATIO`
        """
        plt.figure(figsize=figure_size)

        def get_true_pos_false_pos(threshold):
            temp_eval = TwoClassEvaluator(actual_values=self._actual_values,
                                          predicted_scores=self._predicted_scores,
                                          labels=('x', 'y'),
                                          score_threshold=threshold)

            return threshold, temp_eval.true_positive_rate, temp_eval.false_positive_rate

        auc_curve = [get_true_pos_false_pos(threshold=x) for x in np.arange(0.0, 1.01, 0.01)]
        auc_curve = pd.DataFrame(auc_curve,
                                 columns=['threshold', 'True Positive Rate', 'False Positive Rate'])

        axis = sns.lineplot(data=auc_curve, x='False Positive Rate', y='True Positive Rate', ci=None)
        axis.set_title(f"AUC: {round(self.auc, 3)}")
        for i, (x, y, s) in enumerate(zip(auc_curve['False Positive Rate'],  # pylint: disable=invalid-name
                                          auc_curve['True Positive Rate'],
                                          auc_curve['threshold'])):
            if i % 5 == 0:
                axis.text(x, y, f'{s:.3}')
        axis.set_xticks(np.arange(0, 1.1, .1))
        axis.set_yticks(np.arange(0, 1.1, .1))
        plt.grid()
        plt.tight_layout()

    def plot_threshold_curves(self,
                              score_threshold_range: Tuple[float, float] = (0.3, 0.9),
                              figure_size: tuple = STANDARD_WIDTH_HEIGHT):
        """Plots various scores (e.g. True Positive Rate, False Positive Rate, etc.) for various score
        thresholds. (A score threshold is the value for which you would predict a positive label if the
        value of the score is above the threshold (e.g. usually 0.5).

        Args:
            score_threshold_range:
                range of score thresholds to plot (x-axis); tuple with minimum threshold in first index and
                maximum threshold in second index.
            figure_size:
                tuple containing `(width, height)` of plot. The default height is defined by
                `helpsk.plot.STANDARD_HEIGHT`, and the default width is
                `helpsk.plot.STANDARD_HEIGHT / helpsk.plot.GOLDEN_RATIO`
        """
        plt.figure(figsize=figure_size)

        def get_threshold_scores(threshold):
            temp_eval = TwoClassEvaluator(actual_values=self._actual_values,
                                          predicted_scores=self._predicted_scores,
                                          labels=('x', 'y'),
                                          score_threshold=threshold)

            return threshold,\
                temp_eval.true_positive_rate,\
                temp_eval.false_positive_rate,\
                temp_eval.positive_predictive_value,\
                temp_eval.false_negative_rate,\
                temp_eval.true_negative_rate

        threshold_curves = [get_threshold_scores(threshold=x) for x in np.arange(score_threshold_range[0],
                                                                                 score_threshold_range[1],
                                                                                 0.025)]
        threshold_curves = pd.DataFrame(threshold_curves,
                                        columns=['Score Threshold',
                                                 'True Pos. Rate (Recall)',
                                                 'False Pos. Rate',
                                                 'Pos. Predictive Value (Precision)',
                                                 'False Neg. Rate',
                                                 'True Neg. Rate (Specificity)'])

        axis = sns.lineplot(x='Score Threshold', y='value', hue='variable',
                            data=pd.melt(frame=threshold_curves, id_vars=['Score Threshold']))
        axis.set_xticks(np.arange(score_threshold_range[0], score_threshold_range[1] + 0.1, 0.1))
        axis.set_yticks(np.arange(0, 1.1, .1))
        plt.vlines(x=self.score_threshold, ymin=0, ymax=1, colors='black')
        plt.grid()
        plt.tight_layout()

    def plot_precision_recall_tradeoff(self,
                                       score_threshold_range: Tuple[float, float] = (0, 1),
                                       figure_size: tuple = STANDARD_WIDTH_HEIGHT):
        """Plots the tradeoff between precision (i.e. positive predict value) and recall (i.e. True Positive
        Rate) for various score thresholds. (A score threshold is the value for which you would predict a
        positive label if the value of the score is above the threshold (e.g. usually 0.5).

        Args:
            score_threshold_range:
                range of score thresholds to plot (x-axis); tuple with minimum threshold in first index and
                maximum threshold in second index.
            figure_size:
                tuple containing `(width, height)` of plot. The default height is defined by
                `helpsk.plot.STANDARD_HEIGHT`, and the default width is
                `helpsk.plot.STANDARD_HEIGHT / helpsk.plot.GOLDEN_RATIO`
        """
        plt.figure(figsize=figure_size)

        def get_threshold_scores(threshold):
            temp_eval = TwoClassEvaluator(actual_values=self._actual_values,
                                          predicted_scores=self._predicted_scores,
                                          labels=('x', 'y'),
                                          score_threshold=threshold)

            return threshold,\
                temp_eval.true_positive_rate,\
                temp_eval.positive_predictive_value

        threshold_curves = [get_threshold_scores(threshold=x) for x in np.arange(score_threshold_range[0],
                                                                                 score_threshold_range[1],
                                                                                 0.025)]
        threshold_curves = pd.DataFrame(threshold_curves,
                                        columns=['Score Threshold',
                                                 'True Pos. Rate (Recall)',
                                                 'Pos. Predictive Value (Precision)'])

        axis = sns.lineplot(x='Score Threshold', y='value', hue='variable',
                            data=pd.melt(frame=threshold_curves, id_vars=['Score Threshold']))
        axis.set_xticks(np.arange(score_threshold_range[0], score_threshold_range[1] + 0.1, 0.1))
        axis.set_yticks(np.arange(0, 1.1, .1))
        plt.vlines(x=self.score_threshold, ymin=0, ymax=1, colors='black')
        plt.grid()
        plt.tight_layout()

    def calculate_lift_gain(self,
                            num_buckets: int = 20,
                            return_style: bool = False,
                            include_all_info: bool = False) -> Union[pd.DataFrame, Styler]:
        """
        https://www.listendata.com/2014/08/excel-template-gain-and-lift-charts.html

        Gain is the % of positive (actual) events we have 'captured' i.e. located by looking at the
        top x% of predicted scores, such that the highest scores are looked at first.
        For example, if the percentile is `5%` and the gain value is `0.3`, we can say that if we randomly
        searched `5%` of the data, we would expect to uncover about `5%` of the positive events;
        however, we have uncovered 30% of events by searching the highest 5% of scores.
        Lift is simply the ratio of the percent of events that what we have uncovered for a given percentile
        of data (i.e. gain) divided by what we would have expected by random chance (i.e. the percentile).
        So in the previous example, we uncovered 30% of positive events by searching the top 5% of scores;
        whereas if we took a random sample, we would have only expected to find 5% of the positive events.
        The lift is `0.3 / 0.05` which is `6`; meaning we found `6` times the amount of positive events by
        searching the top 5% of scores, than if we were to randomly sample the data.
        """
        data = pd.DataFrame({
            'predicted_scores': self._predicted_scores,
            'actual_values': self._actual_values,
        })
        data.sort_values(['predicted_scores'], ascending=False, inplace=True)

        # .qcut gets percentiles
        bins = pd.qcut(x=data['predicted_scores'],
                       q=num_buckets,
                       labels=list(range(100, 0, round(-100 / num_buckets))))

        data['Percentile'] = bins

        def gain_grouping(group):
            results = {
                '# of Obs.': len(group.actual_values),
                '# of Pos. Events': sum(group.actual_values == 1)
            }
            return pd.Series(results, index=['# of Obs.', '# of Pos. Events'])

        gain_lift_data = data.groupby('Percentile').apply(gain_grouping)
        temp = pd.DataFrame({'# of Obs.': 0, '# of Pos. Events': 0}, index=[0])
        temp.index.names = ['Percentile']
        gain_lift_data = pd.concat([gain_lift_data, temp])
        gain_lift_data.sort_index(ascending=True, inplace=True)

        gain_lift_data['Cumul. Pos. Events'] = gain_lift_data['# of Pos. Events'].cumsum()
        gain_lift_data['Gain'] = gain_lift_data['Cumul. Pos. Events'] / self._actual_positives
        gain_lift_data = gain_lift_data.loc[~(gain_lift_data.index == 0), :]
        gain_lift_data['Lift'] = gain_lift_data['Gain'] / (gain_lift_data.index.values / 100)

        if not include_all_info:
            gain_lift_data = gain_lift_data[['Gain', 'Lift']]

        gain_lift_data = gain_lift_data.round(2)

        if return_style:
            gain_lift_data = gain_lift_data.style
            gain_lift_data.format(precision=2). \
                bar(subset='Gain', color=hcolor.Colors.PASTEL_BLUE.value,
                    vmin=0, vmax=1). \
                bar(subset='Lift', color=hcolor.Colors.PASTEL_BLUE.value)

        return gain_lift_data


class RegressionEvaluator:
    """
    Evaluates models for regression (i.e. numeric outcome) problems.
    """
    def __init__(self,
                 actual_values: np.ndarray,
                 predicted_values: np.ndarray):
        """
        Args:
            actual_values:
                the actual values
            predicted_values:
                the predicted values
        """

        assert len(actual_values) == len(predicted_values)

        self._actual_values = actual_values
        self._predicted_values = predicted_values
        self._residuals = actual_values - predicted_values
        self._standard_deviation = np.std(actual_values)
        self._mean_squared_error = float(np.mean(np.square(actual_values - predicted_values)))
        self._mean_absolute_error = float(np.mean(np.abs(actual_values - predicted_values)))
        self._r_squared = r2_score(y_true=actual_values, y_pred=predicted_values)

    @property
    def mean_absolute_error(self) -> float:
        """Mean Absolute Error"""
        return self._mean_absolute_error

    @property
    def mean_squared_error(self) -> float:
        """Mean Squared Error"""
        return self._mean_squared_error

    @property
    def root_mean_squared_error(self) -> float:
        """Root Mean Squared Error"""
        return np.sqrt(self.mean_squared_error)

    @property
    def rmse_to_st_dev(self) -> float:
        """The ratio of RMSE to the standard deviation of the actual values.
        Gives an indication of how large the errors are to the actual data.
        """
        return self.root_mean_squared_error / self._standard_deviation

    @property
    def r_squared(self) -> float:
        """R Squared"""
        return self._r_squared

    @property
    def total_observations(self):
        """The total number of observations i.e. sample size."""
        return len(self._actual_values)

    @property
    def all_metrics(self) -> dict:
        """Returns a dictionary of the most common error metrics for regression problems."""
        return {'Mean Absolute Error (MAE)': self.mean_absolute_error,
                'Root Mean Squared Error (RMSE)': self.root_mean_squared_error,
                'RMSE to Standard Deviation of Target': self.rmse_to_st_dev,
                'R Squared': self.r_squared,
                'Total Observations': self.total_observations}

    def all_metrics_df(self,
                       return_style: bool = False,
                       round_by: Optional[int] = None) -> Union[pd.DataFrame, Styler]:
        """All of the metrics are returned as a DataFrame.

        Args:
            return_style:
                if True, return styler object; else return dataframe
            round_by:
                the number of digits to round by; if None, then don't round
        """
        result = pd.DataFrame.from_dict({key: value for key, value in self.all_metrics.items()},  # pylint: disable=unnecessary-comprehension  # noqa
                                        orient='index',
                                        columns=['Scores'])

        if round_by is not None:
            result.iloc[0:2] = result.iloc[0:2].round(round_by)

        if return_style:
            subset_scores = pd.IndexSlice[result.loc[['Mean Absolute Error (MAE)',
                                                      'Root Mean Squared Error (RMSE)'],
                                                     'Scores'].index, 'Scores']
            subset_secondary = pd.IndexSlice[result.loc[['RMSE to Standard Deviation of Target',
                                                         'R Squared'],
                                                        'Scores'].index, 'Scores']
            subset_total_observations = pd.IndexSlice[result.loc[['Total Observations'],
                                                                 'Scores'].index, 'Scores']
            result = result.style
            if round_by is not None:
                result = result.format(subset=subset_scores, thousands=',', precision=round_by)
            else:
                result = result.format(subset=subset_scores, thousands=',')
            result = result.format(subset=subset_secondary, precision=3)
            result = result.format(subset=subset_total_observations, thousands=',', precision=0)

        return result

    def plot_residuals_vs_fits(self, figure_size: tuple = STANDARD_WIDTH_HEIGHT):
        """Plots residuals vs fitted values

        Args:
            figure_size:
                tuple containing `(width, height)` of plot. The default height is defined by
                `helpsk.plot.STANDARD_HEIGHT`, and the default width is
                `helpsk.plot.STANDARD_HEIGHT / helpsk.plot.GOLDEN_RATIO`
        """
        lowess = sm.nonparametric.lowess
        loess_points = lowess(self._residuals, self._predicted_values)
        loess_x, loess_y = zip(*loess_points)

        plt.figure(figsize=figure_size)
        plt.plot(loess_x, loess_y, color='r')
        plt.scatter(x=self._predicted_values, y=self._residuals, s=8, alpha=0.5)
        plt.title('Residuals vs. Fitted Values')
        plt.xlabel('Fitted Values')
        plt.ylabel('Residuals (Actual - Predicted)')

    def plot_predictions_vs_actuals(self, figure_size: tuple = STANDARD_WIDTH_HEIGHT):
        """Plots predictions vs actual values

        Args:
            figure_size:
                tuple containing `(width, height)` of plot. The default height is defined by
                `helpsk.plot.STANDARD_HEIGHT`, and the default width is
                `helpsk.plot.STANDARD_HEIGHT / helpsk.plot.GOLDEN_RATIO`
        """
        lowess = sm.nonparametric.lowess
        loess_points = lowess(self._predicted_values, self._actual_values)
        loess_x, loess_y = zip(*loess_points)

        plt.figure(figsize=figure_size)
        plt.plot(loess_x, loess_y, color='r', alpha=0.5, label='Loess (Predictions vs Actuals)')
        plt.plot(self._actual_values, self._actual_values, color='b', alpha=0.5, label='Perfect Prediction')
        plt.scatter(x=self._actual_values, y=self._predicted_values, s=8, alpha=0.5)
        plt.title('Predicted Values vs. Actual Values')
        plt.xlabel('Actuals')
        plt.ylabel('Predicted')
        axis = plt.gca()
        handles, labels = axis.get_legend_handles_labels()
        axis.legend(handles, labels)
        plt.figtext(0.99, 0.01,
                    'Note: observations above blue line mean model is over-predicting; below means under-'
                    'predicting.',  # noqa
                    horizontalalignment='right')
        return axis

    def plot_residuals_vs_actuals(self, figure_size: tuple = STANDARD_WIDTH_HEIGHT):
        """Plots residuals vs actuals values

        Args:
            figure_size:
                tuple containing `(width, height)` of plot. The default height is defined by
                `helpsk.plot.STANDARD_HEIGHT`, and the default width is
                `helpsk.plot.STANDARD_HEIGHT / helpsk.plot.GOLDEN_RATIO`
        """
        lowess = sm.nonparametric.lowess
        loess_points = lowess(self._residuals, self._actual_values)
        loess_x, loess_y = zip(*loess_points)

        plt.figure(figsize=figure_size)
        plt.plot(loess_x, loess_y, color='r')
        plt.scatter(x=self._actual_values, y=self._residuals, s=8, alpha=0.5)
        plt.title('Residuals vs. Actual Values')
        plt.xlabel('Actual')
        plt.ylabel('Residuals (Actual - Predicted)')
        plt.figtext(0.99, 0.01,
                    'Note: Actual > Predicted => Under-predicting (positive residual); negative residuals '
                    'mean over-predicting',  # noqa
                    horizontalalignment='right')



from datetime import date, timedelta
import json
import cryptocode
import click

@click.command()
@click.option('-m', help='The MAC address of the machine running the software', required=True)
@click.option('-i', help='The system id of the machine running the software', required=True)
@click.option('-s', help='The start date of the license', default=date.today().strftime('%Y%m%d'))
@click.option('-e', help='The end date of the license', default=(date.today() + timedelta(days=365)).strftime('%Y%m%d'))
@click.option('-f', help='The features of the license', default='all')
@click.option('-u', help='The version of the software using the license', default='v1.0.0')
def create_license_file(m, i, s, e, f, u):
    licence = {'system_id': i,
               'mac_address': m,
               'start_date': s,
               'end_date': e,
               'features': f,
               'version': u}

    licence_json = json.dumps(licence)
    encrypt_key = f'2020license{m}'
    encrypted_json = cryptocode.encrypt(licence_json, encrypt_key)

    with open(f'{i}.lic', 'w') as f:
        f.write(encrypted_json)


def main():
    create_license_file()


if __name__ == '__main__':
    main()


"""This model contains classes that wrap the connection/querying logic of various databases. Database objects
can be created from the __init__ function or from the .from_config class function by passing in an instance
of the corresponding configuration class.

Examples:

    with Redshift(user='username', password='asdf', ....) as redshift:
        redshift.query("SELECT * FROM table LIMIT 100")

    or

    redshift_config = RedshiftConfigFile('/path/to/redshift.config')
    with Redshift.from_config(redshift_config) as redshift:
        redshift.query("SELECT * FROM table LIMIT 100")
"""
import configparser
from typing import Union

import pandas as pd

from helpsk.database_base import Configuration, Database


class GenericConfigFile(Configuration):  # pylint: disable=too-few-public-methods
    """Class that is used to map a configuration file in the format below, to a dictionary that will be passed
    into the corresponding Database object.

        [config_key]
        user=[user]
        password=[password]
        database=[database]
        port=[port]
        host=[host]
    """
    def __init__(self, config_file: str, config_key: str, config_mapping: dict):
        """
        Args:
             config_file:
                the path to the configuration file

             config_key:
                the configuration key i.e. keyword of configuration items

             config_mapping:
                a dictionary containing the mapping to go from the configuration item names to the keys of the
                dictionary created. The keys should be the same as the Database object's constructor.

                So if the Database constructor is `__init__(user, password, database)`

                and the configuration file is in the format of:

                    [database_config]
                    user_item=robert
                    password_item=123
                    database_item=my_database

                then the config_mapping dict should be

                    {'user': 'user_item',
                     'password': 'password_item',
                     'database': 'database_item'}

                and the resulting dict from calling get_dictionary would be

                    {'user': 'robert',
                     'password': '123',
                     'database': 'my_database'}

                which would be passed into the Database object, either as

                    config_file = GenericConfigFile(...)
                    Database(**config_file.get_dictionary())

                or

                    config_file = GenericConfigFile(...)
                    Database.from_config(config_file)
        """
        self._config_file = config_file
        self._config_key = config_key
        self._config_mapping = config_mapping

    def get_dictionary(self) -> dict:
        """ logic that builds the dictionary described in the __init__ docstring

        Returns:
             a dictionary to be passed to the Database object
        """
        config = configparser.ConfigParser()
        config.read(self._config_file)
        return {key: config[self._config_key][value] for key, value in self._config_mapping.items()}


class RedshiftConfigFile(GenericConfigFile):  # pylint: disable=too-few-public-methods
    """Supplies a standard config_key and config_mapping to the GenericConfigFile object.

    Corresponds to a configuration file in the format of:

        [redshift]
        user=[user]
        password=[password]
        database=[database]
        port=[port]
        host=[host]
    """
    def __init__(self, config_file: str, config_key: str = 'redshift'):
        """
        Args:
            config_file:
                the path to the configuration file

            config_key:
                the configuration key, i.e. the `[xxx]` part of the configuration file.
                The default value is `redshift`.

        """
        config_mapping = {'user': 'username',
                          'password': 'password',
                          'database': 'database',
                          'port': 'port',
                          'host': 'host'}

        super().__init__(config_file=config_file, config_key=config_key, config_mapping=config_mapping)


class SnowflakeConfigFile(GenericConfigFile):  # pylint: disable=too-few-public-methods
    """Supplies a standard config_key and config_mapping to the GenericConfigFile object.

    Corresponds to a configuration file in the format of:

        [snowflake]
        user=[user]
        account=[account]
        authenticator=[authenticator]
        warehouse=[warehouse]
        database=[database]
    """
    def __init__(self, config_file: str, config_key: str = 'snowflake'):
        """
        Args:
            config_file:
                the path to the configuration file

            config_key:
                the configuration key, i.e. the `[xxx]` part of the configuration file.
                The default value is `snowflake`.

        """
        config_mapping = {'user': 'username',
                          'account': 'account',
                          'authenticator': 'authenticator',
                          'warehouse': 'warehouse',
                          'database': 'database'}

        super().__init__(config_file=config_file, config_key=config_key, config_mapping=config_mapping)


class Redshift(Database):
    """Wraps logic for connecting to Redshift and querying.

    Example:
        config = RedshiftConfigFile('/path/to/redshift.config')
        with Redshift.from_config(config) as redshift:
            redshift.query("SELECT * FROM table LIMIT 100")
    """
    def __init__(self, *,
                 user: str, password: str, database: str, host: str, port: Union[str, int]):  # pylint: disable=too-many-arguments
        """Initialization"""
        super().__init__()
        self._connection_string = "dbname={} host={} port={} user={} password={}". \
            format(database, host, port, user, password)

    def _open_connection_object(self) -> object:
        """Wraps logic for connecting to redshift
        """
        from psycopg2 import connect  # pylint: disable=import-outside-toplevel
        return connect(self._connection_string)

    def _close_connection_object(self):
        """Wraps logic for closing the connection to redshift
        """
        self.connection_object.close()

    def _query(self, sql: str) -> pd.DataFrame:
        """Wraps logic for querying redshift.

        Args:
            sql:
                SQL to execute e.g. "SELECT * FROM table LIMIT 100"

        Returns:
            a pandas Dataframe with the results from the query
        """
        return pd.read_sql_query(sql, self.connection_object)


class Snowflake(Database):
    """Wraps logic for connecting to Snowflake and querying.

        Example:
            config = SnowflakeConfigFile('/path/to/snowflake.config')
            with Snowflake.from_config(config) as snowflake:
                snowflake.query("SELECT * FROM table LIMIT 100")

    Instructions for installing snowflake dependencies:
        https://docs.snowflake.com/en/user-guide/python-connector-install.html

    I used v2.5.0, and python v3.9 (requirements_39):

    ```
    pip install -r ......./snowflake-connector-python/v2.5.0/tested_requirements/requirements_39.reqs
    pip install snowflake-connector-python==v2.5.0
    ```

    Additionally:
        https://docs.snowflake.com/en/user-guide/python-connector-pandas.html#installation

    ```
    pip install snowflake-connector-python[pandas]
    ```
    """
    # pylint: disable=too-many-arguments
    def __init__(self, *,
                 user: str, account: str, authenticator: str, warehouse: str, database: str,
                 autocommit: bool = False):
        """Initialization"""
        super().__init__()
        self.user = user
        self.account = account
        self.authenticator = authenticator
        self.warehouse = warehouse
        self.database = database
        self.autocommit = autocommit

    def _open_connection_object(self) -> object:
        """Wraps logic for connecting to snowflake
        """
        from snowflake.connector import connect  # pylint: disable=import-outside-toplevel
        return connect(
            user=self.user,
            account=self.account,
            authenticator=self.authenticator,
            warehouse=self.warehouse,
            database=self.database,
            autocommit=self.autocommit,
        )

    def _close_connection_object(self):
        """Wraps logic for closing the connection to snowflake
        """
        self.connection_object.close()

    def _query(self, sql: str) -> pd.DataFrame:
        """Wraps logic for querying snowflake.

        Args:
            sql:
                SQL to execute e.g. "SELECT * FROM table LIMIT 100"

        Returns:
            a pandas Dataframe with the results from the query
        """
        cursor = self.connection_object.cursor()
        cursor.execute(sql)
        dataframe = cursor.fetch_pandas_all()

        return dataframe



import unittest
import warnings  # noqa

import numpy as np
import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.impute import SimpleImputer
from sklearn.metrics import make_scorer, f1_score, precision_score, recall_score, SCORERS, roc_auc_score, fbeta_score, cohen_kappa_score, \
    confusion_matrix, mean_squared_error, mean_absolute_error
from sklearn.model_selection import train_test_split, GridSearchCV, RepeatedKFold
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import label_binarize, StandardScaler, OneHotEncoder

import helpsk as hlp
from helpsk.sklearn_eval import SearchCVParser, TwoClassEvaluator, RegressionEvaluator
from helpsk.sklearn_pipeline import CustomOrdinalEncoder
from helpsk.utility import suppress_warnings
from tests.helpers import get_data_credit, get_test_path, check_plot, helper_test_dataframe, get_data_housing


def warn(*args, **kwargs):  # noqa
    pass
warnings.warn = warn  # noqa


# noinspection PyMethodMayBeStatic
class TestSklearnEval(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        ####
        # set up grid-search on classification model for credit data
        ####

        credit_data = get_data_credit()
        credit_data.loc[0:46, ['duration']] = np.nan
        credit_data.loc[25:75, ['checking_status']] = np.nan
        credit_data.loc[10:54, ['credit_amount']] = 0
        y_full = credit_data['target']
        X_full = credit_data.drop(columns='target')  # noqa
        y_full = label_binarize(y_full, classes=['bad', 'good']).flatten()
        X_train, X_test, y_train, y_test = train_test_split(X_full, y_full, test_size=0.2, random_state=42)  # noqa
        del y_full, X_full
        numeric_columns = hlp.pandas.get_numeric_columns(X_train)
        non_numeric_columns = hlp.pandas.get_non_numeric_columns(X_train)
        numeric_pipeline = Pipeline([
            ('imputing', SimpleImputer(strategy='mean')),
            ('scaling', StandardScaler()),
        ])
        non_numeric_pipeline = Pipeline([
            ('encoder_chooser', hlp.sklearn_pipeline.TransformerChooser()),
        ])
        transformations_pipeline = ColumnTransformer([
            ('numeric_pipeline', numeric_pipeline, numeric_columns),
            ('non_numeric_pipeline', non_numeric_pipeline, non_numeric_columns)
        ])
        random_forest_model = RandomForestClassifier(random_state=42)
        full_pipeline = Pipeline([
            ('preparation', transformations_pipeline),
            ('model', random_forest_model)
        ])
        param_grad = [
            {
                'preparation__non_numeric_pipeline__encoder_chooser__transformer': [OneHotEncoder(),
                                                                                    CustomOrdinalEncoder()],
                'model__max_features': [100, 'auto'],
                'model__n_estimators': [10, 50]
            },
        ]
        # https://github.com/scikit-learn/scikit-learn/blob/2beed55847ee70d363bdbfe14ee4401438fba057/sklearn_eval/metrics/_scorer.py#L702
        # https://scikit-learn.org/stable/modules/generated/sklearn.metrics.make_scorer.html
        # https://stackoverflow.com/questions/60615281/different-result-roc-auc-score-and-plot-roc-curve
        scores = {
            # https://github.com/scikit-learn/scikit-learn/blob/2beed5584/sklearn_eval/metrics/_scorer.py#L537
            'ROC/AUC': SCORERS['roc_auc'],
            'F1': make_scorer(f1_score, greater_is_better=True),
            'Pos. Pred. Val': make_scorer(precision_score, greater_is_better=True),
            'True Pos. Rate': make_scorer(recall_score, greater_is_better=True),
        }
        grid_search = GridSearchCV(full_pipeline,
                                   param_grid=param_grad,
                                   cv=RepeatedKFold(n_splits=3, n_repeats=1, random_state=42),
                                   scoring=scores,
                                   refit=False,
                                   return_train_score=True)
        grid_search.fit(X_train, y_train)
        cls.credit_data__grid_search = grid_search

        grid_search = GridSearchCV(full_pipeline,
                                   param_grid=param_grad,
                                   cv=RepeatedKFold(n_splits=3, n_repeats=1, random_state=42),
                                   scoring='roc_auc',
                                   refit=True,
                                   return_train_score=True)
        grid_search.fit(X_train, y_train)
        cls.credit_data__grid_search__roc_auc = grid_search

        best_model = grid_search.best_estimator_
        predicted_scores = best_model.predict_proba(X_test)[:, 1]
        cls.credit_data__y_test = y_test
        cls.credit_data__y_scores = predicted_scores

        ####
        # set up grid-search on regression model for housing data
        ####
        housing_data = get_data_housing()
        housing_data.loc[0:46, ['median_income']] = np.nan
        housing_data.loc[25:75, ['housing_median_age']] = np.nan
        y_full = housing_data['target']
        X_full = housing_data.drop(columns='target')  # noqa
        X_train, X_test, y_train, y_test = train_test_split(X_full, y_full, test_size=0.2, random_state=42)  # noqa
        del y_full, X_full
        numeric_pipeline = Pipeline([
            ('imputing', SimpleImputer(strategy='mean')),
            ('scaling', StandardScaler()),
        ])
        random_forest_model = RandomForestRegressor(random_state=42)
        full_pipeline = Pipeline([
            ('preparation', numeric_pipeline),
            ('model', random_forest_model)
        ])
        param_grad = [
            {
                'model__max_features': [2, 'auto'],
                'model__n_estimators': [10, 50]
            },
        ]
        scores = {
            'RMSE': make_scorer(mean_squared_error, greater_is_better=False, squared=False),
            'MAE': make_scorer(mean_absolute_error, greater_is_better=False),
        }
        grid_search = GridSearchCV(full_pipeline,
                                   param_grid=param_grad,
                                   cv=RepeatedKFold(n_splits=3, n_repeats=1, random_state=42),
                                   scoring=scores,
                                   refit='RMSE',
                                   return_train_score=True)
        grid_search.fit(X_train, y_train)
        cls.housing_data__grid_search = grid_search
        best_model = grid_search.best_estimator_
        predicted_values = best_model.predict(X_test)
        cls.housing_data__y_test = y_test
        cls.housing_data__y_predictions = predicted_values

    def test_cv_results_to_dataframe(self):
        grid_search = self.credit_data__grid_search
        new_param_column_names = {'model | max_features': 'max_features',
                                  'model | n_estimators': 'n_estimators',
                                  'preparation | non_numeric_pipeline | encoder_chooser | transformer':
                                      'encoder'}
        cv_results = SearchCVParser(searcher=grid_search,
                                    higher_score_is_better=True,
                                    new_param_column_names=new_param_column_names)
#
#
#         cv_results.trial_labels
#
#
#
#
#         import plotly.graph_objects as go
#         import plotly.io as pio
#         pio.renderers.default = "browser"
#
#         data = cv_results.results[~cv_results.results.iloc[:, 0].isna()]
#         data = data.iloc[::-1]  # reverse order to display "best" on top of y-axis
#
# #        cv_results.primary_score_standard_errors.loc[data.index]
#
#
#         fig = go.Figure(data=go.Scatter(
#             x=data.iloc[:, 0],
#             y=cv_results.trial_labels.loc[data.index],
#             error_x=dict(
#                 type='data',
#                 symmetric=False,
#                 array=data.iloc[:, 2] - data.iloc[:, 0],
#                 arrayminus=data.iloc[:, 0] - data.iloc[:, 1])
#         ))
#         fig.update_yaxes(type='category')
#         fig.add_vline(x=cv_results.top_score - cv_results.top_score_standard_error, line_dash='dash')
#         fig.update_yaxes(
#             scaleanchor="x",
#             scaleratio=1.618,
#         )
#         fig.show()

        self.assertEqual(cv_results.trial_labels.shape, (cv_results.number_of_trials, ))
        self.assertFalse(cv_results.trial_labels.isna().any())
        self.assertEqual(cv_results.top_score, cv_results.results.iloc[0, 0])
        self.assertEqual(cv_results.result_indexes_within_1_standard_error.tolist(), [6, 7])
        self.assertEqual(cv_results.results.index.tolist(),
                         cv_results.primary_score_standard_errors.index.tolist())
        self.assertEqual(cv_results.primary_score_standard_errors.iloc[0],
                         cv_results.top_score_standard_error)

        self.assertEqual(cv_results.results.index.tolist(), [6, 7, 4, 5, 0, 1, 2, 3])
        self.assertEqual(cv_results.results.index.tolist(),
                         cv_results.formatted_results(return_style=False).index.tolist())

        self.assertEqual(cv_results.score_names, ['ROC/AUC', 'F1', 'Pos. Pred. Val', 'True Pos. Rate'])
        self.assertEqual(cv_results.score_columns,
                         ['ROC/AUC Mean',
                          'ROC/AUC 95CI.LO',
                          'ROC/AUC 95CI.HI',
                          'F1 Mean',
                          'F1 95CI.LO',
                          'F1 95CI.HI',
                          'Pos. Pred. Val Mean',
                          'Pos. Pred. Val 95CI.LO',
                          'Pos. Pred. Val 95CI.HI',
                          'True Pos. Rate Mean',
                          'True Pos. Rate 95CI.LO',
                          'True Pos. Rate 95CI.HI'])
        self.assertEqual(cv_results.training_score_columns,
                         ['ROC/AUC Training Mean',
                          'F1 Training Mean',
                          'Pos. Pred. Val Training Mean',
                          'True Pos. Rate Training Mean'])
        self.assertEqual(cv_results.parameter_columns,
                         ['max_features',
                          'n_estimators',
                          'encoder'])

        self.assertEqual(cv_results.results.shape[0], cv_results.number_of_trials)
        self.assertEqual(cv_results.formatted_results(return_style=False).shape[0], cv_results.number_of_trials)
        self.assertEqual(len(cv_results.fit_time_per_trial), cv_results.number_of_trials)
        self.assertEqual(len(cv_results.score_time_per_trial), cv_results.number_of_trials)

        results = cv_results.formatted_results(return_train_score=True, return_style=False)
        self.assertIsInstance(results, pd.DataFrame)
        self.assertIsInstance(results['encoder'].iloc[0], str)
        equal = results.columns == ['ROC/AUC Mean', 'ROC/AUC 95CI.LO', 'ROC/AUC 95CI.HI',
                                    'ROC/AUC Training Mean',
                                    'F1 Mean', 'F1 95CI.LO', 'F1 95CI.HI',
                                    'F1 Training Mean',
                                    'Pos. Pred. Val Mean', 'Pos. Pred. Val 95CI.LO', 'Pos. Pred. Val 95CI.HI',
                                    'Pos. Pred. Val Training Mean',
                                    'True Pos. Rate Mean', 'True Pos. Rate 95CI.LO', 'True Pos. Rate 95CI.HI',
                                    'True Pos. Rate Training Mean',
                                    'max_features', 'n_estimators',
                                    'encoder']
        self.assertTrue(all(equal))

        results = cv_results.formatted_results(return_train_score=False, return_style=False)

        self.assertIsInstance(results, pd.DataFrame)
        self.assertIsInstance(results['encoder'].iloc[0],
                              str)
        equal = results.columns == ['ROC/AUC Mean', 'ROC/AUC 95CI.LO', 'ROC/AUC 95CI.HI',
                                    'F1 Mean', 'F1 95CI.LO', 'F1 95CI.HI',
                                    'Pos. Pred. Val Mean', 'Pos. Pred. Val 95CI.LO', 'Pos. Pred. Val 95CI.HI',
                                    'True Pos. Rate Mean', 'True Pos. Rate 95CI.LO', 'True Pos. Rate 95CI.HI',
                                    'max_features', 'n_estimators', 'encoder']
        self.assertTrue(all(equal))

        with suppress_warnings():
            results = cv_results.formatted_results(return_train_score=True, return_style=True)
        with open(get_test_path() + '/test_files/sklearn_eval/credit__grid_search__with_training.html', 'w') as file:
            file.write(results.render())

        with suppress_warnings():
            results = cv_results.formatted_results(return_train_score=False, return_style=True)
        with open(get_test_path() + '/test_files/sklearn_eval/credit__grid_search__without_training.html', 'w') as file:
            file.write(results.render())

        grid_search = self.credit_data__grid_search__roc_auc
        cv_results = SearchCVParser(searcher=grid_search, higher_score_is_better=True)

        self.assertEqual(cv_results.trial_labels.shape, (cv_results.number_of_trials,))
        self.assertFalse(cv_results.trial_labels.isna().any())
        self.assertEqual(cv_results.top_score, cv_results.results.iloc[0, 0])
        self.assertEqual(cv_results.result_indexes_within_1_standard_error.tolist(), [6, 7])
        self.assertEqual(cv_results.results.index.tolist(),
                         cv_results.primary_score_standard_errors.index.tolist())
        self.assertEqual(cv_results.primary_score_standard_errors.iloc[0],
                         cv_results.top_score_standard_error)

        self.assertEqual(cv_results.score_names, ['roc_auc'])
        self.assertEqual(cv_results.score_columns,
                         ['roc_auc Mean',
                          'roc_auc 95CI.LO',
                          'roc_auc 95CI.HI'])
        self.assertEqual(cv_results.training_score_columns, ['roc_auc Training Mean'])
        self.assertEqual(cv_results.parameter_columns,
                         ['model | max_features',
                          'model | n_estimators',
                          'preparation | non_numeric_pipeline | encoder_chooser | transformer'])

        self.assertEqual(cv_results.results.shape[0], cv_results.number_of_trials)
        self.assertEqual(cv_results.formatted_results(return_style=False).shape[0], cv_results.number_of_trials)
        self.assertEqual(len(cv_results.fit_time_per_trial), cv_results.number_of_trials)
        self.assertEqual(len(cv_results.score_time_per_trial), cv_results.number_of_trials)

        with suppress_warnings():
            results = cv_results.formatted_results(return_train_score=True, return_style=True)
        test_file = get_test_path() + '/test_files/sklearn_eval/credit__grid_search__default_scores__with_training.html'
        with open(test_file, 'w') as file:
            file.write(results.render())

        with suppress_warnings():
            results = cv_results.formatted_results(return_train_score=False, return_style=True)
        test_file = get_test_path() + '/test_files/sklearn_eval/credit__grid_search__default_scores__without_training.html'
        with open(test_file, 'w') as file:
            file.write(results.render())

    def test_cv_results_to_dataframe_regression(self):
        grid_search = self.housing_data__grid_search
        cv_results = SearchCVParser(searcher=grid_search, higher_score_is_better=False)

        self.assertEqual(cv_results.trial_labels.shape, (cv_results.number_of_trials,))
        self.assertFalse(cv_results.trial_labels.isna().any())
        self.assertEqual(cv_results.top_score, cv_results.results.iloc[0, 0])
        self.assertEqual(cv_results.result_indexes_within_1_standard_error.tolist(), [3])
        self.assertEqual(cv_results.results.index.tolist(),
                         cv_results.primary_score_standard_errors.index.tolist())
        self.assertEqual(cv_results.primary_score_standard_errors.iloc[0],
                         cv_results.top_score_standard_error)

        self.assertEqual(cv_results.score_names, ['RMSE', 'MAE'])
        self.assertEqual(cv_results.score_columns,
                         ['RMSE Mean',
                          'RMSE 95CI.LO',
                          'RMSE 95CI.HI',
                          'MAE Mean',
                          'MAE 95CI.LO',
                          'MAE 95CI.HI'])
        self.assertEqual(cv_results.training_score_columns, ['RMSE Training Mean', 'MAE Training Mean'])
        self.assertEqual(cv_results.parameter_columns, ['model | max_features', 'model | n_estimators'])

        self.assertEqual(cv_results.results.shape[0], cv_results.number_of_trials)
        self.assertEqual(cv_results.formatted_results(return_style=False).shape[0], cv_results.number_of_trials)
        self.assertEqual(len(cv_results.fit_time_per_trial), cv_results.number_of_trials)
        self.assertEqual(len(cv_results.score_time_per_trial), cv_results.number_of_trials)

        results = cv_results.formatted_results(return_train_score=True, return_style=False)

        self.assertIsInstance(results, pd.DataFrame)
        self.assertIsInstance(results['model | max_features'].iloc[0], str)
        equal = results.columns == ['RMSE Mean', 'RMSE 95CI.LO', 'RMSE 95CI.HI', 'RMSE Training Mean',
                                    'MAE Mean', 'MAE 95CI.LO', 'MAE 95CI.HI', 'MAE Training Mean',
                                    'model | max_features', 'model | n_estimators']
        self.assertTrue(all(equal))

        results = cv_results.formatted_results(return_train_score=False, return_style=False)

        self.assertIsInstance(results, pd.DataFrame)
        self.assertIsInstance(results['model | max_features'].iloc[0], str)
        equal = results.columns == ['RMSE Mean', 'RMSE 95CI.LO', 'RMSE 95CI.HI',
                                    'MAE Mean', 'MAE 95CI.LO', 'MAE 95CI.HI',
                                    'model | max_features', 'model | n_estimators']
        self.assertTrue(all(equal))

        with suppress_warnings():
            results = cv_results.formatted_results(return_train_score=True, return_style=True)
        with open(get_test_path() + '/test_files/sklearn_eval/housing__grid_search__with_training.html', 'w') as file:
            file.write(results.render())

        with suppress_warnings():
            results = cv_results.formatted_results(return_train_score=False, return_style=True)
        with open(get_test_path() + '/test_files/sklearn_eval/housing__grid_search__without_training.html', 'w') as file:
            file.write(results.render())

    def test_TwoClassEvaluator(self):
        y_true = self.credit_data__y_test
        y_score = self.credit_data__y_scores
        score_threshold = 0.5
        y_pred = [1 if x > score_threshold else 0 for x in y_score]
        evaluator = TwoClassEvaluator(actual_values=self.credit_data__y_test,
                                      predicted_scores=self.credit_data__y_scores,
                                      labels=('Bad', 'Good'),
                                      score_threshold=0.5)

        con_matrix = confusion_matrix(y_true=y_true, y_pred=y_pred)
        self.assertEqual(evaluator._true_negatives, con_matrix[0, 0])
        self.assertEqual(evaluator._false_positives, con_matrix[0, 1])
        self.assertEqual(evaluator._false_negatives, con_matrix[1, 0])
        self.assertEqual(evaluator._true_positives, con_matrix[1, 1])
        self.assertEqual(evaluator.sample_size, len(self.credit_data__y_test))
        self.assertTrue(evaluator.true_positive_rate == evaluator.sensitivity == evaluator.recall)
        self.assertEqual(evaluator.recall, recall_score(y_true=y_true, y_pred=y_pred))
        self.assertEqual(evaluator.true_negative_rate, evaluator.specificity)
        self.assertEqual(round(evaluator.false_positive_rate, 9), round(1 - evaluator.specificity, 9))
        self.assertEqual(evaluator.positive_predictive_value, evaluator.precision)
        self.assertEqual(evaluator.precision, precision_score(y_true=y_true, y_pred=y_pred))
        self.assertEqual(evaluator.f1_score, f1_score(y_true=y_true, y_pred=y_pred))
        self.assertEqual(evaluator.auc, roc_auc_score(y_true=y_true, y_score=y_score))
        self.assertEqual(evaluator.fbeta_score(beta=0), fbeta_score(y_true=y_true, y_pred=y_pred, beta=0))
        self.assertEqual(evaluator.fbeta_score(beta=1), fbeta_score(y_true=y_true, y_pred=y_pred, beta=1))
        self.assertEqual(evaluator.fbeta_score(beta=1), evaluator.f1_score)
        self.assertEqual(evaluator.fbeta_score(beta=2), fbeta_score(y_true=y_true, y_pred=y_pred, beta=2))
        self.assertEqual(round(evaluator.kappa, 9), round(cohen_kappa_score(y1=y_true, y2=y_pred), 9))

        self.assertIsInstance(evaluator.all_metrics, dict)
        self.assertIsInstance(evaluator.all_metrics_df(return_style=False), pd.DataFrame)

        with open(get_test_path() + '/test_files/sklearn_eval/all_metrics_df.html', 'w') as file:
            table_html = evaluator.all_metrics_df(return_details=False, return_style=True).render()
            file.write(table_html)

        with open(get_test_path() + '/test_files/sklearn_eval/all_metrics_df__round_3.html', 'w') as file:
            table_html = evaluator.all_metrics_df(return_details=False,
                                                  return_style=True,
                                                  round_by=3).render()
            file.write(table_html)

        with open(get_test_path() + '/test_files/sklearn_eval/all_metrics_df__with_details.html', 'w') as file:
            table_html = evaluator.all_metrics_df(return_details=True, return_style=True).render()
            file.write(table_html)

        with open(get_test_path() + '/test_files/sklearn_eval/all_metrics_df__with_details__round_3.html', 'w') as file:
            table_html = evaluator.all_metrics_df(return_details=True,
                                                  return_style=True,
                                                  round_by=3).render()
            file.write(table_html)

    def test_RegressionEvaluator(self):
        evaluator = RegressionEvaluator(actual_values=self.housing_data__y_test,
                                        predicted_values=self.housing_data__y_predictions)

        self.assertEqual(evaluator.mean_absolute_error,
                         mean_absolute_error(y_true=self.housing_data__y_test,
                                             y_pred=self.housing_data__y_predictions))
        self.assertEqual(round(mean_squared_error(y_true=self.housing_data__y_test,
                                                  y_pred=self.housing_data__y_predictions), 4),
                         round(evaluator.mean_squared_error, 4))
        self.assertEqual(np.sqrt(evaluator.mean_squared_error), evaluator.root_mean_squared_error)
        self.assertEqual(evaluator.total_observations, len(self.housing_data__y_test))
        self.assertIsInstance(evaluator.all_metrics, dict)
        self.assertIsInstance(evaluator.all_metrics_df(), pd.DataFrame)

        with open(get_test_path() + '/test_files/sklearn_eval/reg_eval__all_metrics_df.html', 'w') as file:
            table_html = evaluator.all_metrics_df(return_style=True).render()
            file.write(table_html)

        with open(get_test_path() + '/test_files/sklearn_eval/reg_eval__all_metrics_df__round_3.html', 'w') as file:
            table_html = evaluator.all_metrics_df(return_style=True, round_by=3).render()
            file.write(table_html)

        with open(get_test_path() + '/test_files/sklearn_eval/reg_eval__all_metrics_df__round_0.html', 'w') as file:
            table_html = evaluator.all_metrics_df(return_style=True, round_by=0).render()
            file.write(table_html)

        check_plot(file_name=get_test_path() + '/test_files/sklearn_eval/reg_eval__plot_residuals_vs_fits.png',
                   plot_function=lambda: evaluator.plot_residuals_vs_fits())

        check_plot(file_name=get_test_path() + '/test_files/sklearn_eval/reg_eval__plot_residuals_vs_actuals.png',
                   plot_function=lambda: evaluator.plot_residuals_vs_actuals())

        check_plot(file_name=get_test_path() + '/test_files/sklearn_eval/reg_eval__plot_predictions_vs_actuals.png',
                   plot_function=lambda: evaluator.plot_predictions_vs_actuals())

    def test_plot_confusion_matrix(self):
        evaluator = TwoClassEvaluator(actual_values=self.credit_data__y_test,
                                      predicted_scores=self.credit_data__y_scores,
                                      labels=('Bad', 'Good'),
                                      score_threshold=0.5)

        check_plot(file_name=get_test_path() + '/test_files/sklearn_eval/plot_confusion_matrix.png',
                   plot_function=lambda: evaluator.plot_confusion_matrix())

    def test_plot_auc_curve(self):
        evaluator = TwoClassEvaluator(actual_values=self.credit_data__y_test,
                                      predicted_scores=self.credit_data__y_scores,
                                      labels=('Bad', 'Good'),
                                      score_threshold=0.5)

        check_plot(file_name=get_test_path() + '/test_files/sklearn_eval/plot_auc_curve.png',
                   plot_function=lambda: evaluator.plot_auc_curve())

    def test_plot_threshold_curves(self):
        evaluator = TwoClassEvaluator(actual_values=self.credit_data__y_test,
                                      predicted_scores=self.credit_data__y_scores,
                                      labels=('Bad', 'Good'),
                                      score_threshold=0.5)

        check_plot(file_name=get_test_path() + '/test_files/sklearn_eval/plot_threshold_curves.png',
                   plot_function=lambda: evaluator.plot_threshold_curves())

    def test_plot_precision_recall_tradeoff(self):
        evaluator = TwoClassEvaluator(actual_values=self.credit_data__y_test,
                                      predicted_scores=self.credit_data__y_scores,
                                      labels=('Bad', 'Good'),
                                      score_threshold=0.5)

        check_plot(file_name=get_test_path() + '/test_files/sklearn_eval/plot_precision_recall_tradeoff.png',
                   plot_function=lambda: evaluator.plot_precision_recall_tradeoff())

    def test_calculate_lift_gain(self):
        evaluator = TwoClassEvaluator(actual_values=self.credit_data__y_test,
                                      predicted_scores=self.credit_data__y_scores,
                                      labels=('Bad', 'Good'),
                                      score_threshold=0.5)

        helper_test_dataframe(file_name=get_test_path() + '/test_files/sklearn_eval/calculate_lift_gain.txt',
                              dataframe=evaluator.calculate_lift_gain())

        helper_test_dataframe(file_name=get_test_path() + '/test_files/sklearn_eval/calculate_lift_gain__10_buckets.txt',
                              dataframe=evaluator.calculate_lift_gain(num_buckets=10))

        with open(get_test_path() + '/test_files/sklearn_eval/calculate_lift_gain.html', 'w') as file:
            table_html = evaluator.calculate_lift_gain(return_style=True).render()
            file.write(table_html)

        with open(get_test_path() + '/test_files/sklearn_eval/calculate_lift_gain__10_buckets.html', 'w') as file:
            table_html = evaluator.calculate_lift_gain(return_style=True, num_buckets=10).render()
            file.write(table_html)



"""Contains a collection of date related helper functions.
"""
from typing import Union
from enum import unique, Enum, auto
import datetime

import pandas as pd
from dateutil import relativedelta
import numpy as np
from helpsk.validation import any_none_nan


@unique
class Granularity(Enum):
    """
    Valid values for date granularity
    """
    DAY = auto()
    MONTH = auto()
    QUARTER = auto()


# pylint: disable=too-many-return-statements
def floor(value: Union[datetime.datetime, datetime.date, pd.Series],
          granularity: Granularity = Granularity.DAY,
          fiscal_start: int = 1) -> [datetime.date, pd.Series]:

    """"Rounds" the datetime value down (i.e. floor) to the the nearest granularity.

    For example, if `date` is `2021-03-20` and granularity is `Granularity.QUARTER` then `floor()` will return
    `2021-01-01`.

    If the fiscal year starts on November (`fiscal_start is `11`; i.e. the quarter is November,
    December, January) and the date is `2022-01-28` and granularity is `Granularity.QUARTER` then `floor()`
    will return `2021-11-01`.

    Args:
        value:
            a datetime or pd.Series of datetimes

        granularity:
            the granularity to round down to (i.e. DAY, MONTH, QUARTER)

        fiscal_start:
            Only applicable for `Granularity.QUARTER`. The value should be the month index (e.g. Jan is 1,
            Feb, 2, etc.) that corresponds to the first month of the fiscal year.

    Returns:
        date - the date rounded down to the nearest granularity
    """
    if isinstance(value, pd.Series):
        return pd.Series([floor(x, granularity=granularity, fiscal_start=fiscal_start) for x in value],
                         name=value.name)

    if any_none_nan(value):
        return value

    if granularity == Granularity.DAY:
        if isinstance(value, datetime.datetime):
            return value.date()

        return value

    if granularity == Granularity.MONTH:
        if isinstance(value, datetime.datetime):
            return value.replace(day=1).date()

        return value.replace(day=1)

    if granularity == Granularity.QUARTER:
        relative_start_index = ((fiscal_start - 1) % 3) + 1
        current_month_index = ((value.month - 1) % 3) + 1
        months_to_subtract = (((relative_start_index * -1) + current_month_index) % 3)
        months_to_subtract = relativedelta.relativedelta(months=months_to_subtract)
        return floor(value, granularity=Granularity.MONTH) - months_to_subtract

    raise ValueError("Unknown Granularity type")


def fiscal_quarter(value: Union[datetime.datetime, datetime.date],
                   include_year: bool = False,
                   fiscal_start: int = 1) -> float:
    """Returns the fiscal quarter (or year and quarter numeric value) for a given date.

    For example:
        from dateutil.parser import parse

        date.fiscal_quarter(parse('2021-01-15')) == 1
        date.fiscal_quarter(parse('2021-01-15'), include_year=True) == 2021.1


        date.fiscal_quarter(parse('2021-01-15'), fiscal_start=2) == 4
        date.fiscal_quarter(parse('2021-01-15'), include_year=True, fiscal_start=2) == 2021.4
        date.fiscal_quarter(parse('2020-11-15'), include_year=True, fiscal_start=2) == 2021.4

    Logic converted from R's Lubridate package:
        https://github.com/tidyverse/lubridate/blob/master/R/accessors-quarter.r

    Args:
        value:
            a date or datetime

        include_year:
            logical indicating whether or not to include the year
            if `True` then returns a float in the form of year.quarter.
            For example, Q4 of 2021 would be returned as `2021.4`

        fiscal_start:
            numeric indicating the starting month of a fiscal year.
            A value of 1 indicates standard quarters (i.e. starting in January).

    Returns:
        date - the date rounded down to the nearest granularity
    """

    assert 1 <= fiscal_start <= 12

    fiscal_start = (fiscal_start - 1) % 12
    shifted = np.arange(fiscal_start, 11 + fiscal_start + 1) % 12 + 1
    quarters = np.repeat([1, 2, 3, 4], 3)
    match_index = np.where(value.month == shifted)
    assert len(match_index) == 1
    match_index = int(match_index[0])
    quarter = quarters[match_index]

    if include_year:
        if fiscal_start == 0:
            return value.year + (quarter / 10)

        next_year_months = np.arange(fiscal_start + 1, 12 + 1)
        return value.year + (value.month in next_year_months) + (quarter / 10)

    return quarter


def to_string(value: Union[datetime.datetime, datetime.date],
              granularity: Granularity = Granularity.DAY,
              fiscal_start: int = 1) -> str:
    """Converts the date to a string.

    Examples:

        ```
        from dateutil.parser import parse

        to_string(value=parse('2021-01-15'), granularity=Granularity.DAY) == "2021-01-15"
        to_string(value=parse('2021-01-15'), granularity=Granularity.Month) == "2021-Jan"
        to_string(value=parse('2021-01-15'), granularity=Granularity.QUARTER) == "2021-Q1"
        to_string(value=parse('2021-01-15'),
                  granularity=Granularity.QUARTER,
                  fiscal_start=2) == "2021-FQ4"
        ```

    If the fiscal year starts on November (`fiscal_start is `11`; i.e. the quarter is November,
    December, January) and the date is `2022-01-28` and granularity is `Granularity.QUARTER` then `floor()`
    will return `2021-11-01`.

    Args:
        value:
            a datetime

        granularity:
            the granularity to round down to (i.e. DAY, MONTH, QUARTER)

        fiscal_start:
            Only applicable for `Granularity.QUARTER`. The value should be the month index (e.g. Jan is 1,
            Feb, 2, etc.) that corresponds to the first month of the fiscal year.

            If fiscal_start start is any other value than `1` quarters will be abbreviated with `F` to denote
            non-standard / fiscal quarters. For example, "2021-FQ4" is the 4th fiscal quarter of 2021.

    Returns:
        date - the date rounded down to the nearest granularity
    """
    if granularity == Granularity.DAY:
        return value.strftime("%Y-%m-%d")

    if granularity == Granularity.MONTH:
        return value.strftime("%Y-%b")

    if granularity == Granularity.QUARTER:
        if fiscal_start == 1:
            return str(fiscal_quarter(value,
                                      include_year=True,
                                      fiscal_start=fiscal_start)).replace('.', '-Q')

        return str(fiscal_quarter(value,
                                  include_year=True,
                                  fiscal_start=fiscal_start)).replace('.', '-FQ')

    raise TypeError('Unrecognized Granularity')



import unittest
from contextlib import contextmanager
from os import path
from unittest.mock import patch

import pandas as pd

from tests.helpers import get_test_path
from helpsk.database import GenericConfigFile, RedshiftConfigFile, SnowflakeConfigFile, Redshift, Snowflake


@contextmanager
def mock_redshift():
    """This is basically a wrapper around mock.patch to hide the mock logic. i.e. This is a context manager
    that wraps the mock context wrapper
    """
    with patch('psycopg2.connect'), patch('pandas.read_sql_query') as pandas_read_mock:
        pandas_read_mock.return_value = pd.DataFrame({'test': ['test']})
        yield


@contextmanager
def mock_snowflake():
    """This is basically a wrapper around mock.patch to hide the mock logic. i.e. This is a context manager
    that wraps the mock context wrapper
    """
    with patch('snowflake.connector.connect') as mock_snowflake_connector:
        # mock snowflake logic:
        # connect() -> cursor() -> ... -> fetch_pandas_all()
        # actual code
        # self.connection_object = snowflake.connector.connect()
        # cursor = self.connection_object.cursor()
        # cursor.execute(sql)
        # dataframe = cursor.fetch_pandas_all()
        mock_con = mock_snowflake_connector.return_value
        mock_cur = mock_con.cursor.return_value
        mock_cur.fetch_pandas_all.return_value = pd.DataFrame({'test': ['test']})
        yield


# noinspection PyMethodMayBeStatic
class TestDatabase(unittest.TestCase):

    def setUp(self) -> None:
        self.sample_redshift_file = get_test_path() + '/test_files/database/sample_redshift.config'
        self.assertTrue(path.isfile(self.sample_redshift_file))

        self.sample_snowflake_file = get_test_path() + '/test_files/database/sample_snowflake.config'
        self.assertTrue(path.isfile(self.sample_snowflake_file))

    def test_GenericConfigFile(self):
        config_mapping = {'user_param': 'username',
                          'password_param': 'password',
                          'database_param': 'database',
                          'port_param': 'port',
                          'host_param': 'host'}
        config = GenericConfigFile(config_file=self.sample_redshift_file,
                                   config_key='redshift',
                                   config_mapping=config_mapping)
        expected_value = {'user_param': 'my_username',
                          'password_param': 'my-password-123',
                          'database_param': 'the_database',
                          'port_param': '1234',
                          'host_param': 'host.address.redshift.amazonaws.com'}
        self.assertEqual(config.get_dictionary(), expected_value)

    def test_RedshiftConfigFile(self):
        config = RedshiftConfigFile(config_file=self.sample_redshift_file)
        config_dict = config.get_dictionary()
        expected_value = {'user': 'my_username',
                          'password': 'my-password-123',
                          'database': 'the_database',
                          'port': '1234',
                          'host': 'host.address.redshift.amazonaws.com'}
        self.assertEqual(expected_value, config_dict)

        # test that this can be passed into the redshift database object, which implies that config_dict
        # contains the proper keyboard arguments
        Redshift(**config_dict)

    def test_SnowflakeConfigFile(self):
        config = SnowflakeConfigFile(config_file=self.sample_snowflake_file)
        config_dict = config.get_dictionary()
        expected_value = {'user': 'my.email@address.com',
                          'account': 'account.id',
                          'authenticator': 'externalbrowser',
                          'warehouse': 'WAREHOUSE_NAME',
                          'database': 'DATABASE_NAME'}
        self.assertEqual(expected_value, config_dict)

        # test that this can be passed into the snowflake database object, which implies that config_dict
        # contains the proper keyboard arguments
        Snowflake(**config_dict)

    def test_Database(self):
        redshift_config = RedshiftConfigFile(config_file=self.sample_redshift_file)
        snowflake_config = SnowflakeConfigFile(config_file=self.sample_snowflake_file)

        def test_via_constructor(db_obj, db_mock):
            self.assertFalse(db_obj.is_connected())
            self.assertIsNone(db_obj.connection_object)

            # mock connection method so that we can "open" the connection to the database
            with db_mock():
                """db_obj is a Database object that is in a closed state and has the mock objects set up"""
                for _ in range(2):  # test that the same object can be opened/closed multiple times
                    # connection should be closed
                    self.assertFalse(db_obj.is_connected())
                    self.assertIsNone(db_obj.connection_object)

                    # test connecting
                    db_obj.connect()
                    self.assertTrue(db_obj.is_connected())
                    self.assertIsNotNone(db_obj.connection_object)
                    self.assertIsInstance(db_obj.connection_object, unittest.mock.MagicMock)
                    db_obj.connect()  # test that calling open again doesn't not fail

                    # test querying
                    results = db_obj.query("SELECT * FROM doesnt_exist LIMIT 100", show_elapsed_time=False)
                    self.assertIsInstance(results, pd.DataFrame)
                    self.assertEqual(results.iloc[0, 0], 'test')
                    # test connection is still open after querying
                    self.assertTrue(db_obj.is_connected())
                    self.assertIsNotNone(db_obj.connection_object)
                    # test subsequent query
                    results = db_obj.query("SELECT * FROM doesnt_exist LIMIT 100", show_elapsed_time=False)
                    self.assertIsInstance(results, pd.DataFrame)
                    self.assertEqual(results.iloc[0, 0], 'test')

                    # test closing
                    db_obj.close()
                    self.assertFalse(db_obj.is_connected())
                    self.assertIsNone(db_obj.connection_object)

        test_via_constructor(db_obj=Redshift(**redshift_config.get_dictionary()), db_mock=mock_redshift)
        test_via_constructor(db_obj=Redshift.from_config(redshift_config), db_mock=mock_redshift)
        test_via_constructor(db_obj=Snowflake(**snowflake_config.get_dictionary()), db_mock=mock_snowflake)
        test_via_constructor(db_obj=Snowflake.from_config(snowflake_config), db_mock=mock_snowflake)

        def test_context_manager(db_class, db_config, db_mock):
            # test context manager
            with db_mock():
                with db_class.from_config(db_config) as db_object:
                    self.assertTrue(db_object.is_connected())
                    self.assertIsNotNone(db_object.connection_object)
                    self.assertIsInstance(db_object.connection_object, unittest.mock.MagicMock)

                    # test querying
                    results = db_object.query("SELECT * FROM doesnt_exist LIMIT 100", show_elapsed_time=False)
                    self.assertIsInstance(results, pd.DataFrame)
                    self.assertEqual(results.iloc[0, 0], 'test')
                    # test connection is still open after querying
                    self.assertTrue(db_object.is_connected())
                    self.assertIsNotNone(db_object.connection_object)

                self.assertFalse(db_object.is_connected())
                self.assertIsNone(db_object.connection_object)

        test_context_manager(db_class=Redshift, db_config=redshift_config, db_mock=mock_redshift)
        test_context_manager(db_class=Snowflake, db_config=snowflake_config, db_mock=mock_snowflake)

        # tests that if failure to connect (i.e. no mock and connection failure) that the object is not in an
        # open state
        database = Redshift.from_config(redshift_config)
        with self.assertRaises(Exception):
            database.connect()
        self.assertFalse(database.is_connected())
        self.assertIsNone(database.connection_object)



import unittest
from os import remove
from os.path import isfile
from docstring_markdown import build_markdown
from docstring_markdown.build_markdown import is_line_docstring, \
    is_single_line_docstring, is_line_function_definition, is_line_class_definition,\
    calculate_indentation_levels
from tests.helpers import get_test_path


# noinspection PyMethodMayBeStatic
class TestDocstringMarkdown(unittest.TestCase):

    @unittest.skip
    def test_execute_build(self):

        with open(get_test_path() + '/test_files/docstring_markdown/execute_build__expected_documentation.md', 'r') as file:
            expected_contents_with_toc = file.read()

        with open(get_test_path() + '/test_files/docstring_markdown/execute_build__expected_documentation__no_toc.md',
                  'r') as file:
            expected_contents_without_toc = file.read()

        def test_helper(function,
                        test_file_name,
                        expected_contents):
            test_file = get_test_path() + test_file_name
            function()
            self.assertTrue(isfile(test_file))
            with open(test_file, 'r') as file_pointer:
                test_file_contents = file_pointer.read()
            remove(test_file)
            self.assertEqual(expected_contents, test_file_contents)

        test_helper(lambda: build_markdown.execute_build(input_path=get_test_path() + '/../helpsk/',
                                                         output_path=get_test_path() + '/test_files'),
                    test_file_name='/test_files/docstring_markdown/documentation.md',
                    expected_contents=expected_contents_with_toc)

        test_helper(lambda: build_markdown.execute_build(input_path=get_test_path() + '/../helpsk/',
                                                         output_path=get_test_path() + '/test_files',
                                                         output_filename='another_file_name.md'),
                    test_file_name='/test_files/docstring_markdown/another_file_name.md',
                    expected_contents=expected_contents_with_toc)

        test_helper(lambda: build_markdown.execute_build(input_path=get_test_path() + '/../helpsk/',
                                                         output_path=get_test_path() + '/test_files',
                                                         toc_off=True),
                    test_file_name='/test_files/docstring_markdown/documentation.md',
                    expected_contents=expected_contents_without_toc)

        test_helper(lambda: build_markdown.execute_build(input_path=get_test_path() + '/../helpsk/',
                                                         output_path=get_test_path() + '/test_files',
                                                         output_filename='another_file_name.md',
                                                         toc_off=True),
                    test_file_name='/test_files/docstring_markdown/another_file_name.md',
                    expected_contents=expected_contents_without_toc)

    def test_is_line_docstring(self):
        self.assertTrue(is_line_docstring('"""'))
        self.assertTrue(is_line_docstring('"""\n'))
        self.assertTrue(is_line_docstring('"""asdf\n'))

    def test_is_single_line_docstring(self):
        self.assertTrue(is_single_line_docstring('"""doc"""'))
        self.assertFalse(is_single_line_docstring('"""'))
        self.assertFalse(is_single_line_docstring('"""doc'))

    def test_is_line_function_definition(self):
        self.assertTrue(is_line_function_definition('def xxx'))
        self.assertTrue(is_line_function_definition('    def xxx('))
        self.assertTrue(is_line_function_definition('        def xxx(adsf...'))

        self.assertFalse(is_line_function_definition('def __xxx'))
        self.assertFalse(is_line_function_definition('    def __xxx('))
        self.assertFalse(is_line_function_definition('        def __xxx(adsf...'))

        self.assertFalse(is_line_function_definition('class xxx'))
        self.assertFalse(is_line_function_definition('    class xxx('))
        self.assertFalse(is_line_function_definition('        class xxx(adsf...'))

    def test_is_line_class_definition(self):
        self.assertFalse(is_line_class_definition('def xxx'))
        self.assertFalse(is_line_class_definition('    def xxx('))
        self.assertFalse(is_line_class_definition('        def xxx(adsf...'))

        self.assertFalse(is_line_class_definition('def __xxx'))
        self.assertFalse(is_line_class_definition('    def __xxx('))
        self.assertFalse(is_line_class_definition('        def __xxx(adsf...'))

        self.assertTrue(is_line_class_definition('class xxx'))
        self.assertTrue(is_line_class_definition('    class xxx('))
        self.assertTrue(is_line_class_definition('        class xxx(adsf...'))

    def test_calculate_indentation_levels(self):
        self.assertEqual(calculate_indentation_levels(leading_spaces=''), 0)
        self.assertEqual(calculate_indentation_levels(leading_spaces='    '), 1)
        self.assertEqual(calculate_indentation_levels(leading_spaces='        '), 2)
        self.assertEqual(calculate_indentation_levels(leading_spaces='            '), 3)
        self.assertEqual(calculate_indentation_levels(leading_spaces='                '), 4)

        with self.assertRaises(AssertionError):
            calculate_indentation_levels(leading_spaces=' ')
        with self.assertRaises(AssertionError):
            calculate_indentation_levels(leading_spaces='  ')
        with self.assertRaises(AssertionError):
            calculate_indentation_levels(leading_spaces='   ')

        with self.assertRaises(AssertionError):
            calculate_indentation_levels(leading_spaces='     ')
        with self.assertRaises(AssertionError):
            calculate_indentation_levels(leading_spaces='      ')
        with self.assertRaises(AssertionError):
            calculate_indentation_levels(leading_spaces='       ')



"""Collection of Exception classes. Used to make debugging easier for external callers of the API and to
ensure that unit tests that use self.raiseException test for the correct exception types.
"""


class HelpskError(Exception):
    """Base-class for all exceptions raised by this package."""


class HelpskParamTypeError(HelpskError):
    """There was an invalid parameter type detected."""


class HelpskParamValueError(HelpskError):
    """There was an invalid parameter type detected."""


class HelpskAssertionError(HelpskError):
    """The expected condition was not True."""



"""
This script is a command line program that extract's docstrings from Python files and generates documentation
in the form of a markdown file.

See `build` method for example usages.

This was more of a proof of concept than it is actually meant to be used in the package. I'll keep here until
a later time and then refactor accordingly, perhaps into a different repo.
"""

import click
import glob
import re

__author__ = "Shane Kercheval"


def is_line_docstring(line: str) -> bool:
    """Returns True if the line starts with 0 or more spaces and \"\"\"
    """
    return bool(re.search(r'^ *?"""', line))


def is_single_line_docstring(line: str) -> bool:
    """Returns True if the line starts with 0 or more spaces and begins and ends with \"\"\"
    """
    line = line.strip()
    return line.startswith('"""') and line.endswith('"""') and line.count('"""') == 2


def is_line_class_definition(line: str) -> bool:
    return bool(re.search('^( *)class ', line))


def is_line_function_definition(line: str) -> bool:
    """Returns true if the corresponding line (of a python file) is the start of a function definition.

    Excludes functions that start with `__` which indicates a private function.

    Args:
        line: a line in a python file
    """
    return bool(re.search('^( *)def ', line)) and 'def __' not in line


def calculate_indentation_levels(leading_spaces: str) -> int:
    """Returns the number of indentation levels, where a level is 4 spaces. So 0 leading_spaces has a level of
    0; 4 leading spaces has a level of 1, and so on. 


    Args:
        leading_spaces:
            A string containing the leading spaces e.g. '', or '    ', and so on.

    """
    assert leading_spaces.strip() == ''
    assert len(leading_spaces) % 4 == 0
    return int(len(leading_spaces) / 4)


def create_table_of_content_item(leading_spaces, line):
    """Helper function to recreate a Table of Content line item. 
    """
    return f'{leading_spaces}- [{line.strip()}](#{line.strip().replace(" ", "-")})\n'


def execute_build(input_path: str,
                  output_path: str = './',
                  output_filename: str = 'documentation.md',
                  toc_off: bool = False):
    """Main logic for extracting dostrings from python files
    """

    top_level_header = '##'

    if not re.compile(r'/$').search(input_path):
        input_path += '/'

    generated_markdown = []
    generated_table_of_contents = []

    for filename in glob.iglob(input_path + '[!_]*.py', recursive=True):

        def remove(items, items_to_remove):
            """Remove `items_to_remove` from list of `items`"""
            return [item for item in items if item not in items_to_remove]

        friendly_filename = '/'.join(remove(filename.split('/'), ['.', '..']))
        generated_markdown.append(f'{top_level_header} {friendly_filename}\n\n')
        generated_table_of_contents.append(create_table_of_content_item(leading_spaces='',
                                                                        line=friendly_filename))

        with open(filename) as f:
            file_contents = f.readlines()

        # cycle through each line in the Python file
        # if certain patterns are detected (e.g. docstring, class, function) then there may be a nested-while
        # that loops through that section of the file.
        line_number = 0
        while line_number < len(file_contents):
            line = file_contents[line_number]

            if is_line_docstring(line):

                docstring_leading_spaces = re.search(r'(^ *)"""', line).group(1)

                if is_single_line_docstring(line):
                    line = line.removeprefix(docstring_leading_spaces + '"""').removesuffix('"""\n')
                    generated_markdown.append(line + "\n")
                else:
                    line = line.removeprefix(docstring_leading_spaces + '"""')
                    line = line.strip()
                    if line:
                        generated_markdown.append(line + "\n")

                    # until next line is docstring
                    while not is_line_docstring(file_contents[line_number + 1]):
                        line_number += 1
                        line = file_contents[line_number]
                        line = line.removeprefix(docstring_leading_spaces)

                        if line.strip().lower() in ['args:', 'returns:']:
                            line = line.strip().replace(':', '').capitalize()
                            line = f'##### {line}\n\n'

                        generated_markdown.append(line)

                    generated_markdown.append('\n')
                    line_number += 1

            if is_line_class_definition(line):
                leading_spaces = re.search("^( *)(?:def|class)", line).group(1)
                levels = calculate_indentation_levels(leading_spaces)

                line = line.removeprefix(leading_spaces)

                # header
                friendly_name = re.sub(r'([(:]).*', '', line).strip()  # r'(\(|:).*'
                generated_markdown.append(f'\n{top_level_header}{"#" * (levels + 1)} {friendly_name}\n\n')
                toc_line_item = create_table_of_content_item(leading_spaces=leading_spaces + '    ',
                                                             line=friendly_name)
                generated_table_of_contents.append(toc_line_item)

            if is_line_function_definition(line):
                leading_spaces = re.search('^( *)(?:def|class)', line).group(1)
                levels = calculate_indentation_levels(leading_spaces)
                line = line.removeprefix(leading_spaces)

                # header
                friendly_name = re.sub(r'([(:]).*', '', line).strip()
                generated_markdown.append(f'\n{top_level_header}{"#" * (levels + 1)} {friendly_name}\n\n')
                
                generated_markdown.append(f"\n```python\n{line}")

                # until next line is docstring
                while not is_line_docstring(file_contents[line_number + 1]):
                    line_number += 1
                    line = file_contents[line_number]
                    line = line.removeprefix(docstring_leading_spaces)  # noqa
                    generated_markdown.append(line)

                generated_markdown.append("```\n\n")  # end definition code-block

                toc_line_item = create_table_of_content_item(leading_spaces=leading_spaces + '    ',
                                                             line=friendly_name)
                generated_table_of_contents.append(toc_line_item)

            line_number += 1
        generated_markdown.append('\n---\n\n')

    generated_table_of_contents.append('\n\n')

    if not output_path.endswith('/'):
        output_path = output_path + '/'


    next_file_type = 'w'
    if not toc_off:
        with open(output_path + output_filename, 'w') as the_file:
            the_file.writelines(generated_table_of_contents)
        next_file_type = 'a'

    with open(output_path + output_filename, next_file_type) as the_file:
        the_file.writelines(generated_markdown)


@click.group()
def main():
    """
    Converts Python docstrings to a single markdown file (documentation.md)
    """
    pass


@main.command()
@click.argument('input_path')
@click.argument('output_path')
@click.option('-output_filename', help="The name of the output (markdown) file.")
@click.option('-toc_off',
              is_flag=True,
              help='Use this flag to exclude the table of contents from the output.')
def build(input_path, output_path, output_filename, toc_off):
    """
    Command that extract's docstrings from Python files and generates documentation in the form of a markdown
    file.

    Example Usage:

        python3 build_markdown.py build ../helpsk . -output_filename my_docs.md

        python3 build_markdown.py build --help
        python3 build_markdown.py build ../path_to_files
        python3 build_markdown.py build ../path_to_files . -toc_off
        python3 build_markdown.py build ../path_to_files ../path_to_output -output_filename my_docs.md
            -toc_off
    """

    if not output_filename:
        output_filename = 'documentation.md'

    print(f"Path of Python files:        '{input_path}'")
    print(f"Path to place markdown file: '{output_path}'")
    print(f"Markdown file Name:          '{output_filename}'")
    print(f"Include Table of Contents:   {str(not toc_off)}")
    execute_build(input_path, output_path, output_filename, toc_off)


if __name__ == "__main__":
    main()



"""A collection of system and environment related helper functions.
"""
import inspect
import warnings
import sys
import os
from contextlib import contextmanager, redirect_stdout


@contextmanager
def redirect_stdout_to_file(file, mode='w'):
    """ Helper context manager that opens a file and redirects standard output to that file.

    Example:

    ```
    with redirect_stdout_to_file(file_name):
        print_dataframe(dataframe)
    ```

    Args:
        file:
            the name and path of the file to open (argument is passed to `open()`)
        mode:
            the mode of the file e.g. `w` (argument is passed to `open()`)
    """
    with open(file, mode) as handle:
        with redirect_stdout(handle):
            yield


@contextmanager
def suppress_stdout():
    """Suppress Output


    ```
    print("Now you see it")
    with suppress_stdout():
        print("Now you don't")
    ```

    code from: https://stackoverflow.com/questions/2125702/how-to-suppress-console-output-in-python
    """
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout


@contextmanager
def suppress_warnings():
    """Simple Wrapper around warnings.catch_warnings()"""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        yield


def is_debugging():
    """
    Returns:
        Returns True if the environment is currently debugging (e.g. set a breakpoint in PyCharm),
        False if not
    """
    for frame in inspect.stack():
        if frame[1].endswith("pydevd.py"):
            return True

    return False
